const char* vorosketchInfo = 

/**********************************************************************/

"Vorosketch v0.22 BETA by Herman Haverkort, 11 September 2022";

/***********************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of vorosketch.cpp may be available from 
http://herman.haverkort.net.


************************************************************************

WHAT IS VOROSKETCH?

Vorosketch is a simple tool to sketch small Voronoi diagrams as a bitmap. 
Use it wisely. It is not designed for speed or for accuracy: it may miss 
features of the diagram that are narrower than a pixel. But it is easy 
to adapt to different distance measures regardless of whether we 
understand the geometry of the bisectors yet.


************************************************************************

SUPPORTED DISTANCE MEASURES

Below are the supported definitions for d(p,q), where p is any point in 
the plane, and q is a site. To select the distance measure named xxx,
write -m xxx on the command line. One may also select weighted 
combinations, e.g. -m a*xxx+b*yyy combines xxx and yyy with coefficients
a and b.

For point or polyline sites:

  euclidean   minimum euclidean distance from p to any point of q

  squared     square of the above

  manhattan   minimum L1-distance from p to any point of q

For point sites only:

  Lx          (where x is a real number other than zero) minimum 
              Lx-distance from p to q

  min         minimum of distance in x and distance in y (L-inf)

  bbox        product of distance in x and distance in y

  max         maximum of distance in x and distance in y (Linf)

  highway     minimum travel time from p to q when higher speed of
              travel is possible on designated line segments

  karlsruhe   minimum distance from p to q when travelling only along
              rays from the origin and circles centred on the origin

  city        minimum travel time from p to q when speed is linearly
              proportional to the distance from the centre

  azimuth     difference between the azimuth coordinates modulo 2pi with
              respect to the origin

  logradius   logarithm of distance from the origin

  koeln       shorthand for azimuth+logradius: minimum travel time from 
              p to q when travelling only along rays from the origin and 
              circles centred on the origin, while speed is linearly 
              proportional to the distance from the centre

  orbitout    1/|q| - 2/(|q|+|p-q|+|p|)
              this is, modulo constant factors, the minimum kinetic
              energy which we would have to give an object at q in order 
              to reach p, subject to gravitation towards the origin
              (Why? The speed at q in orbit around the origin is 
              proportional to the square root of 1/|q| - 1/a, where a 
              is the major axis of the elliptical orbit. Since p has to 
              be on that orbit, too, we get that 2a is the length of the
              path opfqo, where o is the origin and f is the other focal
              point of the elliptical orbit. To minimise 2a, and thus, 
              the necessary speed at q, we put f on the segment pq.)

  orbitin     1/|p| - 2/(|p|+|q-p|+|q|)
              this is, modulo constant factors, the minimum kinetic
              energy which we would have to give an object at p in order 
              to reach q, subject to gravitation towards the origin
              (note that for any given p, the site q that minimises this
              is simply the site that minimises |q-p| with an additive
              weight of |q|)

For polyline sites only:

  angle       1 / smallest opening angle of any wedge with apex at p 
              that includes q
              this is effectively: how small q looks as seen from p

  detour      the distance to a line segment q1q2 is |q1p|+|pq2|-|q1q2|

  dilation    the distance to a line segment q1q2 is 
              (|q1p|+|pq2|-|q1q2|)/|q1q2|

For rooted-vector sites (points q with direction towards a point q'):

  semi        1 if angle between qq' and qp is smaller than pi/2, 
              infinite otherwise

  secant      secant of angle between qq' and qp (infinite if negative)

  catch       distance to q times secant divided by two

  catch@x     unit circles around the sites move as they grow: whenever 
              the radius increase by x, the centre moves by the 
              specified vector. x must be positive.

  turn        minimum angle between qq' and qp

  leftturn    minimum counterclockwise angle from qq' to qp

  dubins      minimum length of a path from q to p that starts in 
              direction q' and has curvature radius at most |qq'|

To include additional user-defined distance measures from 
userdistances.cpp, compile with the -D INCLUDE_USER_DISTANCES option.


***********************************************************************/
const char* vorosketchFileFormat =

"INPUT AND OUTPUT FORMAT                                                \n\
                                                                        \n\
The sites are read from a text file on standard input that contains a   \n\
sequence of numbers separated by spaces or line breaks.                 \n\
                                                                        \n\
The first number is the number of sites (at most 128). Then there is a  \n\
list of numbers for each site: first the number of points (1 for point  \n\
sites, 2 for rooted-vector sites, 2 or more for polyline sites); then   \n\
for each point the x and y coordinate; and finally the weight of the    \n\
site (for unweighted Voronoi Diagrams, just set the weight to 0).       \n\
                                                                        \n\
If the highway distance is used, the highways must be specified after   \n\
the sites in the same format (with weights interpreted as speed)        \n\
                                                                        \n\
The output is a bitmap in .bmp-format on standard output.               \n";


/***********************************************************************

WEIGHTED DISTANCES

By default, the specified site weights are subtracted from the 
calculated distances. If the command line includes the -d option, then
the distances are divided by the specified weights instead. Thus, we 
can create, among others, the following types of Voronoi diagrams:

additively weighted:         euclidean
multiplicatively weighted:   euclidean -d
power diagram:               squared

Note that definitions in the literature may differ with respect to 
whether weights are added or subtracted, whether they are multiplied or 
divided by, and whether they are squared or not. So be sure to input 
the weights in the file in such a way that you get the intended 
result.


************************************************************************

WHAT IF THERE ARE REGIONS OF NON-ZERO AREA, IN WHICH MULTIPLE SITES ARE 
AT THE SAME DISTANCE?

The *two* sites that are listed in the input file first win. For this 
purpose, the sites have to be at exactly the same calculated distance. 
To account for rounding errors, use the -e c option to regard 
distances as equal if they do not differ more than c times a pixel's
width.

In case of polyline sites, rather than identifying regions that are 
closest to an endpoint shared by two line segments, one may want to 
divide those regions by the angular bisector of those segments. This 
can be achieved with the -s c options, which shortens the first and the
last segment of each site by c times a pixel's width.
 

***********************************************************************/
const char* vorosketchOptions = 

"OPTIONS                                                                \n\
                                                                        \n\
-m xxx   use distance measure xxx. The following measures are supported:\n\
         euclidean (default), squared, manhattan, Lx (for any real      \n\
         number x), min, bbox, max, karlsruhe, city, azimuth, logradius,\n\
         koeln, highway, orbitout, orbitin, angle, detour, dilation,    \n\
         semi, secant, catch, catch@x (for positive x), turn, leftturn, \n\
         dubins.                                                        \n\
         Combinations can be made using standard arithmetic with the    \n\
         symbols ln sq sqrt abs ^ * / + - ~ _ | ( ) fx fy fr sx sy sr,  \n\
         where x~y = abs(x-y), x_y = min(x,y), x|y = max(x,y),          \n\
         (fx,fy) are the coordinates of a point the plane; fr its       \n\
         distance from the origin; sx, sy, sr the coordinates and       \n\
         distance of a point site. For example:                         \n\
         \"semi * (euclidean + turn)\"                                  \n\
         (quotes protect the formula from the shell)                    \n\
-d       divide by weights instead of subtracting them                  \n\
-e c     set distance equality threshold to c (default: 0)              \n\
-s c     shorten polyline sites by c times a pixel's width (default: 0) \n\
-2       render 2nd-order Voronoi diagram (cannot be combined with -e)  \n\
-n       render next-closest-site diagram (cannot be combined with -e)  \n\
-f       render farthest-site Voronoi diagram                           \n\
-a       anonymous: do not draw the sites                               \n\
-b       draw region boundaries in black (default when -c, -p and -z not\n\
         selected)                                                      \n\
-c       draw coloured sites and regions, using 19 colours from Sasha   \n\
         Trubetskoy's list of 20 simple, distinct colours               \n\
-p zzz   draw coloured sites and regions using the palette named zzz,   \n\
         where zzz is one of Trubetskoy-Default, Trubetskoy-Original,   \n\
         Trubetskoy-Modified, Grey, Stone, Flat, Bright                 \n\
-h       try to give adjacent regions contrasting colours (experimental)\n\
-l       try to give adjacent regions similar colours (experimental)    \n\
-x       give regions with few neighbours pale colours (experimental)   \n\
-*       mark all sites with the same symbol in the same colour         \n\
-u       draw distance contour line at standard distance from each site \n\
         (0 for subtractively, 1 for divisively weighted diagrams)      \n\
-o       omit shading of regions within standard distance from their    \n\
         site (in closest-site and next-closest-site diagrams)          \n\
         (even if their are no such regions, -o can save some time)     \n\
-i x     draw contour lines at intervals of x units                     \n\
-g x     shading by distance, 50% brightness at distance x              \n\
-z       shading by sunlight based on distance gradient                 \n\
-t x     blacken areas at distance more than x from their site          \n\
-r n     set output resolution to n x n pixels (default: 2000)          \n\
-+       mark the origin                                                \n\
-v       verbose: show option settings and progress indicator on        \n\
         standard error output                                          \n\
-w x     clip diagram to [-x,x] X [-x,x] (default: [-1,1] X [-1,1])     \n\
-@ n     use n random unweighted sites instead of reading sites from    \n\
         standard input                                                 \n\
-?       show how to use this tool                                      \n";

// unused so far: -j, -k, -q , -y


/***********************************************************************

HEADERS


**********************************************************
* Tunable constants etc.                                 *
*********************************************************/

typedef long double DOUBLE;

const int nrPens = 5;

// stroke widths in pixels; must be even, because strokes 
// are centered on pixel corners, not pixel centres:
const int strokeWidth[nrPens] = { 2, 4, 6, 8, 12 };

const int contourPen = 0;
const int bisectorPen = 1;
const int sitePen = 3;
const int hatchPen = 4;
const int siteOutlinePen = 4;
const int arrowPen = 1;
const int arrowHeadPen = 2;
const int markOutlinePen = 2;
const int softBisectorPen = 1;

// maximum number of sites (arbitrary); this only to 
// catch mistakes in the input that would cause the 
// program to take forever 
const int MAXNRSITES = 20000;
const int MAXRESOLUTION = 20000;

// uncomment the following line for checkerboard instead
// of hatching in shared regions:
#define CHECKERBOARD

// length of progress indicator:
const int progressUnits = 40;

// size of grid for pre-scan
const int cellWidth = 20;


/*********************************************************
* Includes                                               *
*********************************************************/

#include <string>
using std::string;
#include <string.h>
#include <sstream>
using std::stringstream;
#include <iostream>
using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;
using std::endl;
#include <cstdlib>
#include <vector>
#include <queue>
#include <algorithm>
using std::swap;
using std::min;
using std::max;
#include <limits>
#include <cmath>
const DOUBLE INF = std::numeric_limits<DOUBLE>::infinity();


/*********************************************************
* General tools headers                                  *
*********************************************************/


template < class T > 
T** newMatrix(const long m, const long n);

template < class T > 
void deleteMatrix(T** A);

template < class T > 
T*** newMatrix(const long l, const long m, const long n);

template < class T > 
void deleteMatrix(T*** A);

DOUBLE rnd(const DOUBLE atLeast = 0, const DOUBLE lessThan = 1);

bool xyInDifferentBracketsOfz(const int x, const int y, const int z);


/*********************************************************
* Input/output headers                                   *
*********************************************************/

enum ErrorLevel 
{ 
  warning, 
  userError, 
  usageError, 
  internalError 
};

class Complaint 
// an object to which an error message can be witten with <<. 
// when the object is destroyed, the error message is produced
// and, if it is not a mere warning, the program halts.
{
    private:
  static const char* errorLabel[];
  stringstream complaint;
  ErrorLevel errorLevel;
    public:
  Complaint(ErrorLevel _errorLevel = userError);
  Complaint(ErrorLevel _errorLevel, const string& expected, const string& read);
  ~Complaint();
  template <typename T> Complaint& operator<<(const T& t);
};

void makeLowerCase(string& s);

// convert a string to a number, verifying that the string 
// does not contain any further content:
DOUBLE strtofstop(const string& str);
int strtoistop(const string& str);


void reportProgress(DOUBLE x, string description);


class RGB
{
    private:
  int mRed;
  int mGreen;
  int mBlue;
    public:
  RGB(const int _red, const int _green, const int _blue);
  RGB(const double _grey = 1);
  const int red() const;
  const int green() const;
  const int blue() const;
  RGB& operator*=(const RGB& colour);
  RGB operator*(const DOUBLE value) const;
  RGB& operator*=(const DOUBLE value);
  void limit();
  static RGB black;
  static RGB grey;
  static RGB lightGrey;
  static RGB faintGrey;
  static RGB white; 
};


DOUBLE contrast(const RGB& a, const RGB& b);


class Options;
RGB* chooseColours(Options& options, const int nrSites, bool** adjacent);


class BitmapFile {
    private:
  ostream& outputStream;
  const int width;
  const int padding;
  int samplesYetToCome;
  void writeInteger(int i, const int bytes) { 
    outputStream.write((char*) &i, bytes); 
  }
    public:
  BitmapFile(ostream& _outputStream, const int _width, const int height);
  BitmapFile& writePixel(RGB colour);
  ~BitmapFile();
};


/*********************************************************
* Options headers                                        *
*********************************************************/

struct OptionSyntax 
{
  int rank;
  string id;
  bool requiresArgument;
  bool multipleAllowed;
  OptionSyntax(
    int _rank, 
    string _id, 
    bool _requiresValue = false,
    bool _multipleAllowed = false
  );
};


class Distance;


class Options
{
    public:
  Distance* distance;
  bool useHighways;
  bool useCompositeDistance;
  enum { subtractive, divisive } weighting;
  DOUBLE equalityThreshold;
  DOUBLE segmentShortening;
  enum { first, second, nextclosest, farthest } order;  
  bool drawSites;
  bool drawRegionBoundaries;
  int nrColours;
  RGB* colours;
  enum { inputOrder, highContrast, lowContrast } colourAssignment;
  bool fadeByDegree;
  bool unicolourSites;
  bool drawStandardContours;
  DOUBLE contourInterval;
  bool shadeInnerAreas;
  DOUBLE contourDistance;
  DOUBLE halfBrightDistance;
  bool sunlight;
  DOUBLE cutOffDistance;
  bool distanceInfoNeeded;
  int width;
  bool markOrigin;
  DOUBLE windowWidth;
  DOUBLE windowOffset;
  DOUBLE pixelHalfWidth;
  DOUBLE pixelsPerUnit;
  int randomSites;
  bool verbose;

  Options();
  void parse(const int argc, const char* argv[]);
  void report(ostream& s) const;
  bool lineWork() const;
  static void usage();

    private:
  static const OptionSyntax syntax[];
};


/*********************************************************
* Geometry headers                                       *
*********************************************************/

const DOUBLE PI = 3.141592653589793;
const DOUBLE TWOPI = PI * 2.0;
const DOUBLE HALFPI = PI * 0.5;


class Point
{
    private:
  DOUBLE mX;
  DOUBLE mY;
  DOUBLE mR; // used to to store the length, first time it is asked
    public:
  Point(const DOUBLE _x = 0, const DOUBLE _y = 0);
  Point normalise(DOUBLE length = 1.0);
  Point operator+(const Point& p) const;
  Point operator-(const Point& p) const;
  DOUBLE operator*(const Point& p) const;
  Point operator*(const DOUBLE scalar) const;
  bool operator==(const Point& p) const;
  DOUBLE x() const; // x-coordinate
  DOUBLE y() const; // y-coordinate
  DOUBLE r(); // length/distance from origin
  DOUBLE fi() const; // direction from origin
  DOUBLE ySlope() const; // dy/dx
  DOUBLE xSlope() const; // dx/dy
  Point rotLeft() const;
  friend istream& operator>>(istream& is, Point& p);
};


istream& operator>>(istream& is, Point& p);


void getTurnRange(
  Point& centre, DOUBLE halfwidth, 
  Point& origin, Point& direction,
  DOUBLE& lowerBound, DOUBLE& upperBound
);


DOUBLE safeCosXY1WithSun(const DOUBLE x, const DOUBLE y);
// computes cosine of vector (x,y,1) and unit-length sun vector 
// -1 if (x,y,1).v = -inf;
//  0 if (x,y,1).v is not a number;
//  1 if (x,y,1).v =  inf


/*********************************************************
* Site headers                                           *
*********************************************************/

typedef unsigned short SiteTypes;
const SiteTypes BASICPOINTTYPE = 1;
const SiteTypes POLYLINETYPE = 2;
const SiteTypes ROOTEDVECTORTYPE = 4;
const SiteTypes POINTTYPE = BASICPOINTTYPE | ROOTEDVECTORTYPE;
const SiteTypes NOTYPES = 0;
const SiteTypes ALLTYPES = ~0;

struct Site
{
  int size;
  Point* points;  // indexed from 0 to size-1
  Point* vectors; // indexed from 0 to size-1, but using only either
    // vector[0]         to store the direction of a rooted-vector site, or
    // vector[1..size-1] to store the directions of polyline segments
  DOUBLE weight;
  int id;
  Site();
  void generateRandom(const int _size, const DOUBLE scale);
  void generateRandomPolyline();
  void erase();
  void setId(int _id);
  void verifyWeight() const;
  void weighDistance(DOUBLE& distance) const;
  ~Site();
};


istream& operator>>(istream& is, Site& s);

istream& readRootedVector(istream& is, Site& s);


/*********************************************************
* Highway map headers                                    *
*********************************************************/


struct ValueId {
  DOUBLE value;
  int id;
  ValueId(DOUBLE _value, int _id);
  bool operator<(const ValueId& vi) const;
};


class Canvas; 


class HighwayMap
{
    private:
  struct HighwaySection
  {
    Point p[2];
    Point invVector[2]; // 2x2 matrix row by row such that
                        // invVector * (p[1] - p[0]) = (1,0).
    DOUBLE speed;
    DOUBLE cotanOfEntry;
    std::vector<ValueId> ramps;
    HighwaySection(Point& p0, Point& p1, DOUBLE _speed);
    void calculateRamps(DOUBLE* t, Point& q);
    void calculateInterchange(
      HighwaySection& other, DOUBLE& tSelf, DOUBLE& tOther);
    Point rampLocation(const DOUBLE t) const;
  };   
  std::vector<HighwaySection> sections;

  struct HighwayNode
  {
    Point location;
    std::vector<ValueId> neighbours;
    DOUBLE* distances;
    HighwayNode(Point& _location, int nrSites);
    void connect(int nodeId, DOUBLE distance);
  };
  std::vector<HighwayNode> nodes;
  int nrPrimaryNodes;

  struct Cell
  {
    std::vector<int> nodes;
    std::vector<int> sections;
  };
  Cell*** cells; // usage: cells[siteId][gridRow][gridColumn]
  const int gridWidth;
  const DOUBLE cellsPerUnit;

  struct Probe 
  {
    DOUBLE distance;
    int nodeId;
    int siteId;
    Probe(DOUBLE _distance, int _nodeId, int _siteId);
    bool operator<(const Probe& p) const;
  };
  DOUBLE distance(Point& p, HighwayNode& n, int siteId);
  DOUBLE distance(Point& p, HighwaySection& hw, int siteId);

    public:
  HighwayMap(std::vector<Site>& sites, std::vector<Site>& highways);
  DOUBLE distance(Point& p, int siteId);

  // for debugging purposes:
  void showNodes(Canvas& canvas);
};


/*********************************************************
* Distance headers                                       *
*********************************************************/

DOUBLE sqr(const DOUBLE x);


DOUBLE safeAcos(const DOUBLE x);


DOUBLE& keepMin(DOUBLE& x, const DOUBLE y);


DOUBLE& keepMax(DOUBLE& x, const DOUBLE y);


void getSumRange(DOUBLE& alw, DOUBLE& aup, const DOUBLE blw, const DOUBLE bup);
// !! assumes none of alw, aup, blw, bup are NaN !!


class Distance 
{
    public:
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  bool supports(const SiteTypes s) const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
  virtual void verifySite(Site& s) const;
};


class FieldDistance: public Distance
// this class is intended for variables that do not depend on the sites,
// but only on the point in the plane whose region we want to compute
{
    public:
  virtual SiteTypes supportedSites() const; 
};


class ConstantDistance: public FieldDistance
{
  DOUBLE weight;
    public:
  ConstantDistance(const DOUBLE _weight);
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class FieldXDistance: public FieldDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class FieldYDistance: public FieldDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class FieldRDistance: public FieldDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SiteDistance: public Distance
// this class is intended for variables that depend only on the sites,
// not on the point in the plane whose region we want to compute
{
    public:
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};
  

class SiteXDistance: public SiteDistance
{
    public:
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class SiteYDistance: public SiteDistance
{
    public:
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class SiteRDistance: public SiteDistance
{
    public:
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


template < char S >
class BinaryCompositionDistance: public Distance
{
    protected:
  Distance* a;
  Distance* b;
    public:
  BinaryCompositionDistance(Distance* _a, Distance* _b);
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  virtual void verifySite(Site& s) const;
};


class MinDistance: public BinaryCompositionDistance< '_' >
{
    public:
  MinDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class MaxDistance: public BinaryCompositionDistance< '|' >
{
    public:
  MaxDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SumDistance: public BinaryCompositionDistance< '+' >
{
    public:
  SumDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class DifferenceDistance: public BinaryCompositionDistance< '-' >
{
    public:
  DifferenceDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class AbsoluteDifferenceDistance: public BinaryCompositionDistance< '~' >
{
    public:
  AbsoluteDifferenceDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class ProductDistance: public BinaryCompositionDistance< '*' >
{
    public:
  ProductDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class RatioDistance: public BinaryCompositionDistance< '/' >
{
    public:
  RatioDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class PowerDistance: public BinaryCompositionDistance< '^' >
{
    public:
  PowerDistance(Distance* _a, Distance* _b):
    BinaryCompositionDistance(_a, _b) {};
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class UnaryOperatorDistance: public Distance
{
    protected:
  Distance* a;
  const string symbol;
    public:
  UnaryOperatorDistance(Distance* _a, const char* _symbol);
  virtual const string description() const;
  virtual SiteTypes supportedSites() const; 
  virtual void verifySite(Site& s) const;
};


class LogDistance: public UnaryOperatorDistance
{
    public:
  LogDistance(Distance* _a);
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class AbsDistance: public UnaryOperatorDistance
{
    public:
  AbsDistance(Distance* _a);
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SquareDistance: public UnaryOperatorDistance
{
    public:
  SquareDistance(Distance* _a);
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SquareRootDistance: public UnaryOperatorDistance
{
    public:
  SquareRootDistance(Distance* _a);
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class PointDistance: public Distance
{
    public:
  virtual SiteTypes supportedSites() const;
};

   
class PolylineDistance: public Distance
{
    public:
  virtual SiteTypes supportedSites() const;
};

   
class SetDistance: public Distance
{
    public:
  virtual SiteTypes supportedSites() const;
};

   
class RootedVectorDistance: public Distance
{
    public:
  virtual SiteTypes supportedSites() const; 
};


class EuclideanDistance: public SetDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SquaredEuclideanDistance: public SetDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class HighwayDistance: public PointDistance
{
    public:
  HighwayDistance();
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class ManhattanDistance: public SetDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class LMinDistance: public PointDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class VolumeDistance: public PointDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class LMaxDistance: public PointDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class LDistance: public PointDistance
{
    private:
  DOUBLE n;
  DOUBLE invn;
    public:
  LDistance(const DOUBLE _n);
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class KarlsruheDistance: public PointDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class AzimuthDistance: public PointDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class AngleDistance: public PolylineDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class DetourDistance: public PolylineDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class DilationDistance: public PolylineDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class SemiDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class SecantDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class CatchDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class MixedCatchDistance: public RootedVectorDistance
{
    private:
  DOUBLE speed;
    public:
  MixedCatchDistance(DOUBLE _speed);
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
};


class TurnDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class LeftTurnDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


class DubinsDistance: public RootedVectorDistance
{
    public:
  virtual const string description() const;
  virtual DOUBLE operator()(Point& p, const Site& s) const;
  virtual void getRange(
    Point& centre, DOUBLE halfwidth, const Site& s,
    DOUBLE& lowerBound, DOUBLE& upperBound
  ) const;
};


Distance* getDistanceObject(string name);


/*********************************************************
* Canvas headers                                         *
*********************************************************/

class Canvas
{
    private:
  const int nrRows;
  const int nrCols;
  RGB* pixels;
  RGB** rows;
  int penSize[nrPens];
  int* penMasks[nrPens];
 
    public:
  Canvas(const int _nrRows, const int _nrCols);
  const RGB& operator()(const int row, const int col) const;
  template < class T > void shade(const int row, const int col, const T& value);
  void draw(const int row, const int col, const int pen, const RGB& colour);
  template < class T > void shade(const int row, const int col, const int pen, const T& value);
  void draw(Point& p, Point& q, const int pen, const RGB& colour);
};


void drawSite(Canvas& c, const Site& s, const int pen, const RGB& colour);


/*********************************************************
* Diagram headers                                        *
*********************************************************/

struct Element
{
  short site;
  DOUBLE shade;
  short contour;
  Element();
  static const short softEdge;
  static const short minorContour;
  static const short majorContour;
  static const short bisector;
};  


void putContour(const short level, Element& e1);
void putContour(const short level, Element& e1, Element& e2);


/*********************************************************
* Owner header                                           *
*********************************************************/

struct Owner // (Region ID)
{
  int primary;     // primary site number;   nrSites if none
  int secondary;   // secondary site number; nrSites if none
  Owner();
  const bool operator==(const Owner& o) const;
  const bool operator!=(const Owner& o) const;
};


template < class T >
bool same(const T& a, const T& b, const T& c, const T& d);


/*********************************************************
* Sample header                                           *
*********************************************************/

struct Sample 
{
  Owner owner;
  DOUBLE* distance;
  int* layer;
  DOUBLE sunlight;
};


/***********************************************************************

IMPLEMENTATION


**********************************************************
* General tools                                          *
*********************************************************/

template < class T > 
T** newMatrix(const long m, const long n)
{
  T** A = new T*[m];
  A[0] = new T[m*n];
  for (int i = 1; i < m; ++i) A[i] = A[i-1] + n;
  return A;
}


template < class T > 
void deleteMatrix(T** A)
{
  delete [] A[0];
  delete [] A;
}


template < class T > 
T*** newMatrix(const long l, const long m, const long n)
{
  T*** A = new T**[l];
  A[0] = new T*[l*m];
  for (int i = 1; i < l; ++i) A[i] = A[i-1] + m;
  A[0][0] = new T[l*m*n];
  int lm = l*m;
  for (int i = 1; i < lm; ++i) A[0][i] = A[0][i-1] + n;
  return A;
}


template < class T > 
void deleteMatrix(T*** A)
{
  delete [] A[0][0];
  delete [] A[0];
  delete [] A;
}


DOUBLE rnd(const DOUBLE atLeast, const DOUBLE lessThan)
{
  return atLeast + ((DOUBLE) rand() / RAND_MAX) * (lessThan - atLeast);
}


bool xyInDifferentBracketsOfz(const int x, const int y, const int z)
{
  if ((x < 0) != (y < 0)) return true;
  if (x < 0) return (x + 1) / z != (y + 1) / z;
  return x / z != y / z;
}


/*********************************************************
* Input/Output                                           *
*********************************************************/

const char* Complaint::errorLabel[4] = 
{ 
  "WARNING", 
  "Error",
  "Error", 
  "Internal error" 
};


Complaint::Complaint(ErrorLevel _errorLevel): 
  errorLevel(_errorLevel) 
{
};

  
Complaint::Complaint(
  ErrorLevel _errorLevel, 
  const string& expected, 
  const string& read
):
  errorLevel(_errorLevel)
{
  complaint << "expected \"" << expected << "\", but read \"" << read << "\".";
}


Complaint::~Complaint()
{
  cerr << endl << errorLabel[(int) errorLevel] << ": " << complaint.str() << endl;
  if (errorLevel == usageError) cerr << "Run vorosketch -? for usage." << endl;
  cerr << endl;
  if (errorLevel > warning) exit((int) errorLevel); 
}


template <typename T> Complaint& Complaint::operator<<(const T& t)
{
  complaint << t; return *this;
}


void makeLowerCase(string& s)
{
  for (int i = 0; i < s.length(); ++i) s[i] = tolower(s[i]);
}


DOUBLE strtofstop(const string& str)
{
  stringstream s(str);
  s >> std::ws;
  if (s.eof()) Complaint(userError)
    << "expected a floating-point number but read nothing.";
  DOUBLE x;
  s >> x >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected a floating-point number but read \"" << str << "\".";
  return x; 
}


int strtoistop(const string& str)
{
  stringstream s(str);
  int n;
  s >> n >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected an integer but read \"" << str << "\".";
  return n; 
}


void reportProgress(DOUBLE x, string description)
{
  if (x >= 1)
    cerr << '\r' << description << ": done." << string(progressUnits - 3, ' ');
  else
  {
    cerr << '\r' << description << ": [";
    for (int i = 1; i < progressUnits; ++i)
    {
      if (x * progressUnits >= i)
        cerr << (i % 5 ? '-' : '+');
      else
        cerr << (i % 5 ? ' ' : ':');
    }
    cerr << "]";
  }
  cerr.flush();
}


RGB::RGB(const int _red, const int _green, const int _blue):
  mRed(_red),
  mGreen(_green),
  mBlue(_blue)
{};


RGB::RGB(const double _grey):
  mRed(_grey * 255),
  mGreen(_grey * 255),
  mBlue(_grey * 255)
{};


const int RGB::red() const
{
  return mRed;
}


const int RGB::green() const
{
  return mGreen;
}


const int RGB::blue() const
{
  return mBlue;
}


RGB& RGB::operator*=(const RGB& colour)
{
  mRed *= colour.mRed; mRed /= 255;
  mGreen *= colour.mGreen; mGreen /= 255;
  mBlue *= colour.mBlue; mBlue /= 255;
  return *this;
}


RGB RGB::operator*(const DOUBLE value) const
{
  return RGB(mRed * value, mGreen * value, mBlue * value);
}


RGB& RGB::operator*=(const DOUBLE value)
{
  mRed *= value;
  mGreen *= value;
  mBlue *= value;
  return *this;
}


void RGB::limit()
{
  mRed = max(0, min(255, mRed));
  mGreen = max(0, min(255, mGreen));
  mBlue = max(0, min(255, mBlue));
}


RGB RGB::black(0,0,0);
RGB RGB::grey(192,192,192);
RGB RGB::lightGrey(224,224,224);
RGB RGB::faintGrey(249,249,249);
RGB RGB::white(255,255,255);


const char* colourSchemes = "\
  trubetskoy-default 19 \
  230  25  75   60 180  75  255 225  25    0 130 200 \
  245 130  48  145  30 180   70 240 240  240  50 230 \
  210 245  60  250 190 212    0 128 128  220 190 225 \
  170 110  40  255 250 200  128   0   0  170 255 195 \
  128 128   0  255 215 180  128 128 128 \
  trubetskoy-original 20 \
  230  25  75   60 180  75  255 225  25    0 130 200 \
  245 130  48  145  30 180   70 240 240  240  50 230 \
  210 245  60  250 190 212    0 128 128  220 190 225 \
  170 110  40  255 250 200  128   0   0  170 255 195 \
  128 128   0  255 215 180    0   0 128  128 128 128 \
  trubetskoy-modified 20 \
  255  25 125   60 180  75  255 225  25    0 130 200 \
  245 130  48  145  30 180   70 240 240  240  50 230 \
  210 245  60  250 190 212    0 128 128  220 190 225 \
  170 110  40  255 250 200  192   0   0  170 255 195 \
  128 128   0  255 215 180   60  60 255  128 128 128 \
  grey 6 \
  128 128 128  179 179 179   77  77  77  153 153 153 \
  102 102 102  217 217 217 \
  stone 20 \
  136 112 112  187 163 163   93  69  69  161 137 137 \
  136 136 112  187 187 163   93  93  69  161 161 137 \
  112 112 136  163 163 187   69  69  93  137 137 161 \
  112 136 136  163 187 187   69  93  93  137 161 161 \
  136 112 136  187 163 187   93  69  93  161 137 161 \
  flat 12 \
  255  28  83   79 238  99  172 201  49    0 147 226 \
  234 124  46  168  35 208   54 186 186  208  40 184 \
  178 135 151  168 164 132  126 189 145   70  70 255 \
  bright 12 \
  255 255   0    0 192 255  255   0 128    0 255  96 \
  160   0 255  255 144   0  208   0 255    0 128 255 \
  255 192   0  255   0 255  192 255   0    0 255 255 \
";


DOUBLE contrast(const RGB& a, const RGB& b)
{
  return 
    abs(sqr(a.red()) - sqr(b.red())) +
    abs(sqr(a.green()) - sqr(b.green())) +
    abs(sqr(a.blue()) - sqr(b.blue()));
}


RGB* chooseColours(Options& options, const int nrSites, bool** adjacent)
{
  // initial colouring:
  int colourId[nrSites];
  for (int i = 0; i < nrSites; ++i) colourId[i] = i % options.nrColours;

  if (options.colourAssignment != Options::inputOrder)
  {
    // maximum allowed iterations for cubic-time improvement round:
    int maxIterations = 1 + (1L << 38) / ((long) nrSites * nrSites * nrSites);
/*
    if (maxIterations == 0) Complaint(warning)
      << "Skipping optimisation of colour assignment: too many sites";
    else
*/
    {
      DOUBLE** diff = newMatrix<DOUBLE>(options.nrColours, options.nrColours);
      for (int i = 0; i < options.nrColours; ++i)
        for (int j = 0; j <= i; ++j)
        {
          diff[i][j] = contrast(options.colours[i], options.colours[j]);
          diff[j][i] = diff[i][j];
        }

      // try to improve:
      bool improved = true;
      for (int iteration = 0; improved && (iteration < maxIterations); ++iteration)
      {
        improved = false;
        stringstream progressLabel; 
        progressLabel << "> Swapping colours (round " << iteration+1 << ')';

        // try swapping two colours:
        for (int i = nrSites-1; i > 0; --i)
        {
          if (options.verbose)
            reportProgress(((DOUBLE) nrSites-i)/nrSites, progressLabel.str());
          for (int j = i-1; j >= 0; --j)
          {
            if (options.colourAssignment == Options::highContrast)
            {
              // try to improve contrast:
              DOUBLE before = INF;
              DOUBLE after = INF;
              for (int k = 0; k < nrSites; ++k)
              {
                if (k == i || k == j) continue;
                if (adjacent[i][k]) 
                { 
                  keepMin(before, diff[colourId[i]][colourId[k]]);
                  keepMin(after,  diff[colourId[j]][colourId[k]]);
                }
                if (adjacent[j][k]) 
                {
                  keepMin(before, diff[colourId[j]][colourId[k]]);
                  keepMin(after,  diff[colourId[i]][colourId[k]]);
                }
              }
              if (after <= before) continue;
            }
            else if (options.colourAssignment == Options::lowContrast)
            {  
              // try to reduce contrast:
              DOUBLE before = 0;
              DOUBLE after = 0;
              for (int k = 0; k < nrSites; ++k)
              {
                if (k == i || k == j) continue;
                if (adjacent[i][k]) 
                { 
                  before += diff[colourId[i]][colourId[k]];
                  after  += diff[colourId[j]][colourId[k]];
                }
                if (adjacent[j][k]) 
                {
                  before += diff[colourId[j]][colourId[k]];
                  after  += diff[colourId[i]][colourId[k]];
                }
              }
              if (after >= before) continue;
            }
            else Complaint(internalError) << "unknown colour optimisation mode";
            swap(colourId[i], colourId[j]);
            improved = true;
            break;
          }
        }
        if (options.verbose)
          reportProgress(1, progressLabel.str());
      }
      deleteMatrix<DOUBLE>(diff);
    }
  }

  RGB* colour = new RGB[nrSites+1];
  for (int i = 0; i < nrSites; ++i) colour[i] = options.colours[colourId[i]];
  colour[nrSites] = RGB::black;

  if (options.fadeByDegree)
  {
    for (int i = 0; i < nrSites; ++i)
    {
      long degree = 0;
      for (int j = 0; j < nrSites; ++j)
        if (adjacent[i][j]) ++degree;
      degree = sqr(sqr(degree));
      DOUBLE w = 1.5 / (1.0 + 2500.0/degree);
      colour[i] = RGB(
        w * colour[i].red()   + (1.0 - w) * 255,
        w * colour[i].green() + (1.0 - w) * 255,
        w * colour[i].blue()  + (1.0 - w) * 255
      );
    }
  }

  return colour;
}


BitmapFile::BitmapFile(ostream& _outputStream, const int _width, const int height):
  outputStream(_outputStream),
  width(_width),
  samplesYetToCome(_width * height),
  padding((4 - ((3*width) % 4)) % 4) 
{
  // write header:
  const int headerSize = 26;
  const int lineSize = width * 3 + padding;
  const int fileSize = headerSize + height * lineSize;
  writeInteger(0x42, 1);
  writeInteger(0x4D, 1);
  writeInteger(fileSize, 4);
  writeInteger(0, 4);
  writeInteger(headerSize, 4);
  writeInteger(12, 4);
  writeInteger(width, 2);
  writeInteger(height, 2);
  writeInteger(1, 2);
  writeInteger(24, 2);
}


BitmapFile& BitmapFile::writePixel(RGB colour) 
{
  if (--samplesYetToCome < 0)
    Complaint(internalError) << "attempt to write to finished bitmap file";
  colour.limit();
  writeInteger(colour.blue(), 1);
  writeInteger(colour.green(), 1);
  writeInteger(colour.red(), 1);
  if (samplesYetToCome % width == 0) writeInteger(0, padding);
  return *this;
}


BitmapFile::~BitmapFile() 
{ 
  if (samplesYetToCome > 0)
    Complaint(internalError) << "bitmap file incomplete";
}


/*********************************************************
* Options                                                *
*********************************************************/

// first declare some helpers:

const OptionSyntax* findOption(const OptionSyntax options[], string argv);

struct Argument 
// class to store options with arguments;
// different options must have unique ranks and ids
{
  int position; // position in the input   
  const OptionSyntax* key;
  string value;
  Argument(int _position, const OptionSyntax& _key, const string& _value);

  // functions to extract a number from the string value, and 
  // verify that the number is within given bounds:
  int intValue(const int min, const int max) const;
  DOUBLE floatValue(const DOUBLE min, const DOUBLE max) const;

  // comparison operator to sort by rank, and for options of equal rank,
  // by position in the input:
  bool operator<(const Argument& a) const;
};


// and here is the implementation:

OptionSyntax::OptionSyntax(
  int _rank, 
  string _id, 
  bool _requiresValue,
  bool _multipleAllowed
):
  rank(_rank), 
  id(_id), 
  requiresArgument(_requiresValue),
  multipleAllowed(_multipleAllowed)
{
};


// Options::Options(const int argc, const char* argv[]):
Options::Options():
  distance(new EuclideanDistance()),
  useHighways(false),
  useCompositeDistance(false),
  weighting(Options::subtractive),
  equalityThreshold(0),
  segmentShortening(0),
  order(Options::first),
  drawSites(true),
  drawRegionBoundaries(false),
  nrColours(1),
  colours(&RGB::white),
  colourAssignment(Options::inputOrder),
  fadeByDegree(false),
  unicolourSites(true),
  drawStandardContours(false),
  contourInterval(0),
  shadeInnerAreas(true),
  contourDistance(0),
  halfBrightDistance(-1),
  sunlight(false),
  cutOffDistance(INF),
  distanceInfoNeeded(true),
  width(2000),
  markOrigin(false),
  windowWidth(2),
  windowOffset(-1),
  pixelHalfWidth(0.0005),
  pixelsPerUnit(1000),
  randomSites(0),
  verbose(false)
{};


void Options::parse(const int argc, const char* argv[])
{
  // collect command line arguments, check syntax
  std::vector< Argument > arguments;
  for (int i = 1; i < argc; ++i)
  {
    string s(argv[i]);
    const OptionSyntax* o = findOption(syntax, s);
    if (!o) 
    {
      if (s[0] != '-')
        Complaint(usageError) << "unrecognised command line parameter \"" << argv[i] << "\".";

      // maybe it is multiple options after a single dash, first parse all but the last one:
      for (; s.length() > 2; s.erase(1,1))
      {
        o = findOption(syntax, s.substr(0,2));
        if (!o) 
          Complaint(usageError) << "unrecognised option \"" << s.substr(0,2) << "\".";
        if (o->requiresArgument)
          Complaint(usageError) << "missing value for \"" << o->id << "\" option.";
        arguments.push_back(Argument(arguments.size(), *o, ""));
      }

      // now identify the last option
      o = findOption(syntax, s);
      if (!o) Complaint(usageError) << "unrecognised option \"" << s << "\".";
    }
    if (o->requiresArgument)
    {
      if (++i == argc) 
        Complaint(usageError) << "missing value for \"" << o->id << "\" option.";
      arguments.push_back(Argument(arguments.size(), *o, argv[i]));
    }
    else
      arguments.push_back(Argument(arguments.size(), *o, ""));
  }

  // sort command line arguments, check for missing or double arguments
  std::sort(arguments.begin(), arguments.end());
  for (int i = 1; i < arguments.size(); ++i)
    if (
      arguments[i].key->rank == arguments[i-1].key->rank && 
      arguments[i].value != arguments[i-1].value &&
      !arguments[i].key->multipleAllowed
    )
      Complaint(usageError) << 
      "multiple values for \"" << arguments[i].key->id + "\" option (\"" << 
      arguments[i-1].value << "\" and \"" <<
      arguments[i].value << "\").";
 
  // process command line arguments
  for (std::vector< Argument >::iterator a = arguments.begin(); 
    a != arguments.end(); 
    ++a)
  {
    switch(a->key->id[1]) 
    {
      case 'm':
        delete distance;
        {
          string name(a->value);
          makeLowerCase(name);
          distance = getDistanceObject(name);
          if (!distance->supportedSites()) Complaint(userError) 
            << "Distance function is not compatible with or has not been implemented "
            << "for any type of site";
        }
        break;
      case 'd': 
        weighting = divisive; 
        contourDistance = 1; 
        break;
      case '2': order = second; shadeInnerAreas = false; break;
      case 'n': order = nextclosest; shadeInnerAreas = false; break;
      case 'f': order = farthest; shadeInnerAreas = false; break;
      case 'e': 
        if (order == second) 
          Complaint(warning) 
            << "-e option not supported for second-order Voronoi Diagrams.";
        else if (order == nextclosest) 
          Complaint(warning) 
            << "-e option not supported for next-closest Voronoi Diagrams.";
        else equalityThreshold = a->floatValue(0, 1) * 2.0 / width;
        break;
      case 's': 
        if (!distance->supports(POLYLINETYPE))
          Complaint(warning) << "-s option irrelevant or not supported for " 
            << distance->description() << " distance.";
        else segmentShortening = a->floatValue(0, 1) * 2.0 / width;
        break;
      case 'a': drawSites = false; break;
      case 'b': drawRegionBoundaries = true; break;
      case 'c': 
        a->value = "Trubetskoy-Default";
        // no break, continue as with -p option:
      case 'p': 
        {
          string name(a->value);
          makeLowerCase(name);
          stringstream s(colourSchemes);
          string header;
          do s >> header; while (s && header != name);
          if (header != name)
            Complaint(usageError) << "unrecognised colour palette name \"" << name << "\"";
          s >> nrColours;
          colours = new RGB[nrColours];
          for (int i = 0; i < nrColours; ++i) 
          {
            int r; int g; int b; s >> r >> g >> b; colours[i] = RGB(r,g,b);
          }
          unicolourSites = false;
        }
        break;
      case 'h': colourAssignment = highContrast; break;
      case 'l': colourAssignment = lowContrast; break;
      case 'x': fadeByDegree = true; break;
      case '*': unicolourSites = true; break;
      case 'o': shadeInnerAreas = false; break;
      case 'u': drawStandardContours = true; break;
      case 'i': contourInterval = a->floatValue(0, 100); break;
      case 'g': 
        halfBrightDistance = a->floatValue(0, 100); 
        shadeInnerAreas = false; 
        break;
      case 'z': 
        if (order == second) 
          Complaint(warning) 
            << "-z option not supported for second-order Voronoi Diagrams.";
        sunlight = true; 
        break;
      case 't': cutOffDistance = a->floatValue(0, 1000); break;
      case 'r': width = a->intValue(2, MAXRESOLUTION); break;
      case 'v': verbose = true; break;
      case '?': usage(); break;
      case '+': markOrigin = true; break;
      case 'w': windowWidth = a->floatValue(0.001, 1000) * 2; break;
      case '@': randomSites = a->intValue(1, MAXNRSITES); break;
      default: Complaint(internalError) << "unrecognized option";
    } 
  }
  if (nrColours == 1 && !sunlight) drawRegionBoundaries = true;
  windowOffset = -windowWidth / 2;
  pixelHalfWidth = windowWidth / width / 2;
  pixelsPerUnit = width / windowWidth;
  distanceInfoNeeded = (
    drawStandardContours ||
    contourInterval > 0 ||
    shadeInnerAreas ||
    halfBrightDistance > 0 ||
    sunlight ||
    cutOffDistance < INF
  );
  if (verbose) report(cerr);
}


void Options::report(ostream& s) const
{
  cerr << "Order: ";
  if (order == first) cerr << "1 (regions determined by closest site)" << endl; 
  else if (order == second) cerr << "2 (regions determined by closest two sites)" << endl; 
  else if (order == nextclosest) cerr << "regions determined by second-closest site" << endl; 
  else if (order == farthest) cerr << "n-1 (regions determined by farthest site)" << endl; 
  else Complaint(internalError) << "order set incorrectly";

  cerr << "Distance measure: " 
    << distance->description() << endl;

  cerr << "Weighting: ";
  if (weighting == subtractive)   cerr << "subtract weights" << endl; 
  else if (weighting == divisive) cerr << "divide by weights" << endl; 
  else Complaint(internalError) << "weighting method set incorrectly";

  cerr << "Distances considered equal if difference at most: " 
    << equalityThreshold << endl;

  if (distance->supports(POLYLINETYPE)) cerr << "Polyline sites shortened by: "
    << segmentShortening << endl;

  cerr << "Image size: " << width << 'x' << width << " pixels" << endl;

  cerr << "Window: [" << windowOffset << ',' << windowWidth + windowOffset
       << "]x[" << windowOffset << ',' << windowWidth + windowOffset << "]" << endl;
  cerr << "Origin: ";
  if (!markOrigin) cerr << "not ";
  cerr << "marked" << endl;

  cerr << "Regions: ";
  if (nrColours == 1) cerr << "white" << endl; 
  else if (nrColours > 1) cerr << "in up to " << nrColours << " colours" << endl; 
  else Complaint(internalError) << "colour scheme set incorrectly";

  cerr << "Sites: " ;
  if (drawSites == false) cerr << "not drawn" << endl;
  else if (nrColours == 1 || unicolourSites) cerr << "all the same" << endl; 
  else cerr << "in matching colours" << endl; 
  
  if (nrColours > 1)
  {
    cerr << "Colour assignment: ";
    if (colourAssignment == inputOrder) cerr << "round-robin in input order";
    else if (colourAssignment == highContrast) 
      cerr << "maximise contrast between adjacent regions";
    else if (colourAssignment == lowContrast)
      cerr << "minimise contrast between adjacent regions";
    else Complaint(internalError) << "colour assignment set incorrectly";
    if (fadeByDegree)
      cerr << "; fade regions with few neighbours" << endl;
    cerr << endl;
  }

  cerr << "Unit distance contours: ";
  if (drawStandardContours)
  {
    cerr << "distance " << contourDistance << " marked by contour line";
    if (shadeInnerAreas) cerr << ", area within shaded";
    cerr << endl;
  }
  else if (shadeInnerAreas) 
    cerr << "area within distance " << contourDistance << " shaded" << endl;
  else cerr << "not shown" << endl;

  cerr << "Outer contour: ";
  if (cutOffDistance < INF)
    cerr << "area outside distance " << cutOffDistance << " blackened" << endl;
  else cerr << "none" << endl;

  cerr << "Further contour lines: ";
  if (contourInterval > 0) 
    cerr << "at " << contourInterval << " intervals" << endl;
  else cerr << "none" << endl;

  cerr << "Shading: ";
  if (sunlight) cerr << "by distance gradient";
  if (halfBrightDistance != -1)
  {
    if (sunlight) cerr << ", "; 
    cerr << "50% brightness at distance " << halfBrightDistance << endl;
  }
  else if (shadeInnerAreas)
  {
    if (sunlight) cerr << ", "; 
    cerr << "dark shade within distance " << contourDistance << endl;
  }
  else 
  {
    if (!sunlight) cerr << "none";
    cerr << endl;
  }
} 


void Options::usage() 
{
  cerr << vorosketchInfo << endl;
#ifdef INCLUDE_USER_DISTANCES
  cerr << "extended with user-defined distance measures" << endl;
#endif
  cerr << endl
       << "USAGE    vorosketch [<options>]" << endl
       << "INPUT    sites in file format described below" << endl
       << "OUTPUT   bmp image file" << endl 
       << endl
       << vorosketchOptions 
       << endl
       << vorosketchFileFormat
       << endl;
  exit(-1);
}


bool Options::lineWork() const
{
  return drawRegionBoundaries || drawStandardContours || contourInterval > 0;
}


const OptionSyntax Options::syntax[28] = 
{
  OptionSyntax( 0, "-?"),
  OptionSyntax( 1, "-i", true),
  OptionSyntax( 2, "-m", true),
  OptionSyntax( 3, "-d"),
  OptionSyntax( 4, "-2"),
  OptionSyntax( 5, "-n"),
  OptionSyntax( 6 ,"-f"),
  OptionSyntax( 7, "-r", true),
  OptionSyntax( 8, "-e", true),
  OptionSyntax( 9, "-s", true),
  OptionSyntax(10, "-c"),
  OptionSyntax(11, "-p", true),
  OptionSyntax(12, "-h"),
  OptionSyntax(13, "-l"),
  OptionSyntax(14, "-x"),
  OptionSyntax(15, "-*"),
  OptionSyntax(16, "-a"),
  OptionSyntax(17, "-b"),
  OptionSyntax(18, "-o"),
  OptionSyntax(19, "-u"),
  OptionSyntax(20, "-g", true),
  OptionSyntax(21, "-z"),
  OptionSyntax(22, "-t", true),
  OptionSyntax(23, "-+"),
  OptionSyntax(24, "-w", true),
  OptionSyntax(25, "-@", true),
  OptionSyntax(98, "-v"),
  OptionSyntax(99, "") // sentinel end
}; 


const OptionSyntax* findOption(const OptionSyntax options[], string argv)
{
  string lowerCaseArg(argv.substr(0,4)); makeLowerCase(lowerCaseArg);
  for (const OptionSyntax* o = &options[0]; ; ++o) 
  {
    if (o->id.empty()) return 0;
    if (lowerCaseArg == o->id.substr(0,4)) return o;
  }
}


Argument::Argument(int _position, const OptionSyntax& _key, const string& _value):
  position(_position), key(&_key), value(_value)
{};

 
int Argument::intValue(const int min, const int max) const
{
  int x = strtoistop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}


DOUBLE Argument::floatValue(const DOUBLE min, const DOUBLE max) const
{
  DOUBLE x = strtofstop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}


bool Argument::operator<(const Argument& a) const 
{ 
  if (key->rank < a.key->rank) return true;
  if (a.key->rank < key->rank) return false;
  return (position < a.position);
}


Options options;


/*********************************************************
* Geometry                                               *
*********************************************************/

Point::Point(const DOUBLE _x, const DOUBLE _y):
  mX(_x), mY(_y), mR(-1)
{};


Point Point::normalise(DOUBLE length)
{
  if (length < 0) 
    Complaint(internalError) << "normalise used with negative length";
  DOUBLE factor = length / r();
  return Point(mX * factor, mY * factor);
}


Point Point::operator+(const Point& p) const
{
  return Point(mX + p.mX, mY + p.mY);
}


Point Point::operator-(const Point& p) const
{
  return Point(mX - p.mX, mY - p.mY);
}


DOUBLE Point::operator*(const Point& p) const
{
  return mX * p.mX + mY * p.mY;
}


Point Point::operator*(const DOUBLE scalar) const
{
  return Point(mX * scalar, mY * scalar);
}


bool Point::operator==(const Point& p) const
{
  return (mX == p.mX && mY == p.mY);
}


DOUBLE Point::x() const
{
  return mX;
}


DOUBLE Point::y() const
{
  return mY;
}


DOUBLE Point::r()
{
  if (mR < 0) mR = sqrt(mX * mX + mY * mY);
  return mR;  
}


DOUBLE Point::fi() const
{
  if (mX == 0 && mY == 0) return 0;
  return atan2(mY, mX);  
}


DOUBLE Point::ySlope() const
{
  return mY/mX;
}


DOUBLE Point::xSlope() const
{
  return mX/mY;
}


Point Point::rotLeft() const
{
  return Point(-mY,mX);
}


istream& operator>>(istream& is, Point& p)
{
  is >> p.mX >> p.mY;
  p.mR = -1;
  return is;
}


void getTurnRange(
  Point& centre, DOUBLE halfwidth, 
  Point& origin, Point& direction,
  DOUBLE& lowerBound, DOUBLE& upperBound
)
{
  if (halfwidth == 0)
    Complaint(internalError) << "calling getTurnRange with halfwidth=0";

  if (direction.x() == 0 && direction.y() == 0) 
  {
    lowerBound = 0;
    upperBound = PI;
    return;
  }

  // if the ray cuts the square, the lower bound is 0
  // if the opposite ray cuts the square, the upper bound is PI
  // otherwise the lower and upper bounds are attained at corners

  // to determine if the ray cuts the square: use left turn distance
  // the ray cuts the square if lower and upper bound are >= PI apart

  DOUBLE fwdLowerBound = TWOPI;
  DOUBLE fwdUpperBound = 0;
  DOUBLE revLowerBound = TWOPI;
  DOUBLE revUpperBound = 0;

  Point corner[4] = {
    centre + Point(-halfwidth, -halfwidth) - origin,
    centre + Point(-halfwidth,  halfwidth) - origin,
    centre + Point( halfwidth, -halfwidth) - origin,
    centre + Point( halfwidth,  halfwidth) - origin
  };
  for (int i = 0; i < 4; ++i)
  {
    if (corner[i].x() == 0 && corner[i].y() == 0) continue;

    DOUBLE angle = 0;
    // rotate such that direction is rotated onto positive x-axis
    // (the transformation also scales, but that does not matter)
    Point n(direction * corner[i], direction.rotLeft() * corner[i]);
    angle = atan2(n.y(), n.x());

    // angle lies between -pi and pi
    keepMin(revLowerBound, angle + PI);
    keepMax(revUpperBound, angle + PI);
    if (angle < 0) angle += TWOPI;
    keepMin(fwdLowerBound, angle);
    keepMax(fwdUpperBound, angle);
  }

  lowerBound = (
    (fwdUpperBound - fwdLowerBound >= PI) ? 
    0 :
    min(fwdLowerBound, TWOPI - fwdUpperBound)
  );
  upperBound = (
    (revUpperBound - revLowerBound >= PI) ?
    PI :
    min(fwdUpperBound, TWOPI - fwdLowerBound)
  );
}


DOUBLE safeCosXY1WithSun(const DOUBLE x, const DOUBLE y)
// computes cosine of vector (x,y,1) and unit-length sun vector 
// -1 if (x,y,1).v = -inf;
//  0 if (x,y,1).v is not a number;
//  1 if (x,y,1).v =  inf
{
  static const DOUBLE sun[3] = {-1.0/3.0, 2.0/3.0, 2.0/3.0};
  DOUBLE dotProduct = (x * sun[0] + y * sun[1] + sun[2]);
  return 
    dotProduct == -INF ? -1 :
    isnan(dotProduct) ?   0 :
    dotProduct ==  INF ?  1 :
    dotProduct / sqrt(sqr(x) + sqr(y) + 1);
}


/*********************************************************
* Sites                                                  *
*********************************************************/

Site::Site():
  size(0),
  weight(0)
{};


void Site::generateRandom(const int _size, const DOUBLE scale)
{
  erase();
  size = _size;
  points = new Point[size];
  vectors = new Point[size];
  DOUBLE length = rnd(0, scale);
  points[0] = Point(
    (2.0 - length) * rnd(-0.5, 0.5),
    (2.0 - length) * rnd(-0.5, 0.5)
  );
  DOUBLE t = rnd(0, TWOPI);
  vectors[size-1] = Point(length * cos(t), length * sin(t));
  if (size == 2)
  {
    vectors[0] = Point(0,0);
    points[1] = points[0] + vectors[1];
  }
  weight = options.contourDistance;
}


void Site::generateRandomPolyline()
{
  erase();
  Point endpoint[2];
  DOUBLE t = rnd(0, TWOPI);
  for (int i = 0; i < 2; ++i)
  {
    DOUBLE dt = rnd(1.5, PI);
    t += dt;
    DOUBLE maxsqr = 0.98 / max(sqr(cos(t)), sqr(sin(t)));
    DOUBLE r = sqrt(rnd(0.5, maxsqr));
    endpoint[i] = Point(r * cos(t), r * sin(t));
  }
  Point travel = endpoint[1] - endpoint[0];
  int nrControlPoints = 2 + travel.r() / 0.3;
  Point wiggle = travel.rotLeft() * 0.15;
  
  Point controlPoint[nrControlPoints];
  controlPoint[0] = endpoint[0];
  for (int i = 1; i < nrControlPoints-1; ++i)
    controlPoint[i] = 
      endpoint[0] * ((DOUBLE) (nrControlPoints - 1 - i) / (nrControlPoints - 1)) +
      endpoint[1] * ((DOUBLE) i / (nrControlPoints - 1)) + 
      wiggle * rnd(-0.5, 0.5);
  controlPoint[nrControlPoints-1] = endpoint[1];

  size = 2 * nrControlPoints - 2;
  points = new Point[size];
  points[0] = endpoint[0];
  for (int i = 1; i < nrControlPoints-1; ++i)
  {
    points[2*i-1] = controlPoint[i-1] * 0.3 + controlPoint[i] * 0.7;
    points[2*i]   = controlPoint[i] * 0.7 + controlPoint[i+1] * 0.3;
  }
  points[size-1] = endpoint[1];

  vectors = new Point[size];
  vectors[0] = Point(0,0); 
  for (int i = 1; i < size; ++i)
    vectors[i] = points[i] - points[i-1];
  weight = options.contourDistance;
}


void Site::erase()
{
  if (size > 0) {
    delete [] points;
    delete [] vectors;
  }
  size = 0;
  weight = 0;
}


void Site::setId(int _id)
{
  id = _id;
}


void Site::verifyWeight() const
{
  static bool zeroWeightWarningGiven = false;
  if (options.weighting == Options::divisive)
  {
    if (weight == 0)
    {
      if (!zeroWeightWarningGiven) 
      {
        Complaint(warning) << "sites with zero weight have empty regions";
        zeroWeightWarningGiven = true;
      }
    } 
    else if (weight < 0)
      Complaint(userError) << "negative weight detected in input";
  }
}


void Site::weighDistance(DOUBLE& distance) const
{
  // assume weights cannot be negative, so we never flip the sign 
  // of infinite distances; 
  if (distance == INF) return;
/* no use?
  if (distance < 0) Complaint(userError) 
    << "negative unweighted distance " << distance << " detected";
*/
  switch(options.weighting)
  {
    case Options::subtractive:
      distance -= weight;
      return;
    case Options::divisive:
/* no use?
      if (weight == 0)
        // assuming distances cannot be negative, in which case we'd get -INF:
        distance = INF;
      else 
*/
      distance /= weight;
      return;
  }
}


Site::~Site()
{
  erase();
}


istream& operator>>(istream& is, Site& s)
{
  s.erase();
  is >> s.size;
  if (s.size < 1) Complaint(userError) << "site without points detected in input";
  s.points = new Point[s.size];
  s.vectors = new Point[s.size];
  is >> s.points[0];
  int nextPoint = 1;
  for (int i = 1; i < s.size; ++i) 
  {
    is >> s.points[nextPoint];
    if (s.points[nextPoint] == s.points[nextPoint-1]) 
      Complaint(warning) << "ignoring repeated point in site or highway";
    else ++nextPoint;
  }
  s.size = nextPoint;
  if (s.size > 1 && options.segmentShortening != 0)
  {
    Point cutFromStart = 
      (s.points[0] - s.points[1]).normalise(options.segmentShortening);
    Point cutFromEnd   = 
      (s.points[s.size-1] - s.points[s.size-2]).normalise(options.segmentShortening);
    s.points[0]        = s.points[0]        - cutFromStart;
    s.points[s.size-1] = s.points[s.size-1] - cutFromEnd;
  }
  s.vectors[0] = Point(0,0);
  for (int i = 1; i < s.size; ++i)
    s.vectors[i] = s.points[i] - s.points[i-1];
  is >> s.weight;
  s.verifyWeight();
  return is;
}


istream& readRootedVector(istream& is, Site& s)
{
  s.erase();
  int n; is >> n; if (n != 2) Complaint(userError) 
    << "expected two, not " << n << " point(s) to define a rooted vector";
  s.size = 1;
  s.points = new Point[1];
  s.vectors = new Point[1];
  is >> s.points[0];
  is >> s.vectors[0];
  s.vectors[0] = s.vectors[0] - s.points[0];
  is >> s.weight;
  s.verifyWeight();
  return is;
}


/*********************************************************
* Highway map                                            *
*********************************************************/


ValueId::ValueId(DOUBLE _value, int _id):
  value(_value),
  id(_id)
{};


bool ValueId::operator<(const ValueId& vi) const
{
  if (value < vi.value) return true;
  if (value > vi.value) return false;
  return (id < vi.id);
}


HighwayMap::HighwaySection::HighwaySection(Point& p0, Point& p1, DOUBLE _speed):
  speed(_speed),
  cotanOfEntry(1.0L / sqrt(sqr(_speed) - 1))
{
  p[0] = p0; 
  p[1] = p1;
  Point vector = p[1]-p[0];
  DOUBLE invSqLength = 1.0L / (vector * vector);
  invVector[0] = vector * invSqLength;
  invVector[1] = vector.rotLeft() * invSqLength;
}


void HighwayMap::HighwaySection::calculateRamps(DOUBLE* t, Point& q)
{
  Point pq = q - p[0]; 
  Point relativeLocation(invVector[0] * pq, invVector[1] * pq); 
  t[0] = relativeLocation.x() - abs(relativeLocation.y()) * cotanOfEntry;
  t[1] = relativeLocation.x() + abs(relativeLocation.y()) * cotanOfEntry;
}


void HighwayMap::HighwaySection::calculateInterchange(
  HighwayMap::HighwaySection& other, DOUBLE& tSelf, DOUBLE& tOther
)
{  
  tSelf = INF; tOther = INF;
  Point otherTranslated = other.p[0] - p[0];
  Point other0relative(invVector[0] * otherTranslated, invVector[1] * otherTranslated);
        otherTranslated = other.p[1] - p[0];
  Point other1relative(invVector[0] * otherTranslated, invVector[1] * otherTranslated);
  if (other0relative.y() == other1relative.y()) return;
  tOther = other0relative.y() / (other0relative.y() - other1relative.y());
  tSelf = other0relative.x() * (-tOther + 1.0) + other1relative.x() * tOther;
}


Point HighwayMap::HighwaySection::rampLocation(const DOUBLE t) const
{
  return p[0]*(-t+1.0) + p[1]*t;
}


HighwayMap::HighwayNode::HighwayNode(Point& _location, const int nrSites):
  location(_location),
  distances(new DOUBLE[nrSites])
{
  for (int i = 0; i < nrSites; ++i) distances[i] = INF;
}


void HighwayMap::HighwayNode::connect(int nodeId, DOUBLE distance)
{
  neighbours.push_back(ValueId(distance, nodeId));
}


HighwayMap::Probe::Probe(DOUBLE _distance, int _nodeId, int _siteId):
  distance(_distance),
  nodeId(_nodeId),
  siteId(_siteId)
{};


bool HighwayMap::Probe::operator<(const HighwayMap::Probe& p) const
{
  if (distance > p.distance) return true;
  if (distance < p.distance) return false;
  return (siteId > p.siteId);
}


// distance via node
DOUBLE HighwayMap::distance(Point& p, HighwayNode& n, int siteId)
{
  return((p - n.location).r() + n.distances[siteId]);
}


// distance via highway section
DOUBLE HighwayMap::distance(Point& p, HighwaySection& hw, int siteId)
{
  DOUBLE d = INF;
  DOUBLE t[2];
  hw.calculateRamps(t, p);
  if (t[0] <= 0) 
    keepMin(d,
      (p - hw.p[0]).r() + 
      nodes[hw.ramps.front().id].distances[siteId]
    );
  if (t[1] >= 1)
    keepMin(d,
      (p - hw.p[1]).r() + 
      nodes[hw.ramps.back().id].distances[siteId]
    );
  for (int i = 0; i < 2; ++i)
  {
    if (t[i] <= 0 || t[i] >= 1) continue;
    Point rampLocation = hw.rampLocation(t[i]);

    int j = 0;
    // find the first node in direction i; will simple linear search do?
    // while (hw.ramps[j].value < t[i]) ++j; // safe bcs last ramp is at 1 >= t[i];

    // alternatively: use binary search:
    int maxj = hw.ramps.size()-1; // size() >= 2, so maxj >= 1
    while (j < maxj-1) // so maxj >= j+2
    {
      int midj = (j + maxj)/2;
      if (hw.ramps[midj].value < t[i]) j = midj; 
      else maxj = midj;
    } 
    // now ramps[j].value < t[i] <= ramps[j+1].value
    ++j;
    // end of binary search

    // now ramps[j-1].value < t[i] <= ramps[j].value
    HighwayNode& n = nodes[hw.ramps[j-1+i].id];
    keepMin(d, 
      (p - rampLocation).r() + 
      (rampLocation - n.location).r() / hw.speed +
      n.distances[siteId]
    );
  }
  return d;
}


HighwayMap::HighwayMap(std::vector<Site>& sites, std::vector<Site>& highways):
  gridWidth((options.width - 1) / cellWidth + 1),
  cellsPerUnit((DOUBLE) options.width / options.windowWidth / cellWidth)
{
  int nrSites = sites.size();
  int nrHighways = highways.size();
  long nrEdges = 0;

  std::priority_queue<Probe> queue;
  // create highway sections; create nodes for their endpoints, and
  // connect them to the other sections' endpoints
  for (int i = 0; i < nrHighways; ++i)
  {
    if (highways[i].size <= 1)
      Complaint(userError) << "highway with only one point detected";
    if (highways[i].weight <= 1)
      Complaint(userError) << "highways must have speed greater than 1";
    int nrNodesOnPreviousHighways = nodes.size();
    for (int j = 0; j < highways[i].size; ++j)
    {
      int h = nodes.size();
      nodes.push_back(HighwayNode(highways[i].points[j], nrSites));
      for (int g = (j == 0 ? h-1 : h-2); g >= 0; --g)
      {
        DOUBLE distance = (nodes[h].location - nodes[g].location).r();
        nodes[h].connect(g, distance); ++nrEdges;
        nodes[g].connect(h, distance); ++nrEdges;
      }
      if (j == 0) continue;
      sections.push_back(
        HighwaySection(
          highways[i].points[j-1], highways[i].points[j], highways[i].weight
        )
      );
      sections.back().ramps.push_back(ValueId(0, h-1));
      sections.back().ramps.push_back(ValueId(1, h  ));
    }
  }
  int nrHighwayEndpoints = nodes.size();

  // create vertices for the sites, connect them to the section
  // endpoints, and insert them in the queue for the SP computation
  for (int s = 0; s < nrSites; ++s)
  {
    queue.push(Probe(0, nrHighwayEndpoints + s, s));
    nodes.push_back(HighwayNode(sites[s].points[0], nrSites));
    for (int g = 0; g < nrHighwayEndpoints; ++g)
      nodes.back().connect(g, (nodes.back().location - nodes[g].location).r());
    nrEdges += nrHighwayEndpoints;
  }
  nrPrimaryNodes = nodes.size();

  // create vertices for all highway intersections
  for (int h = 0; h < sections.size(); ++h)
  {
    HighwaySection& hs = sections[h];
    for (int g = 0; g < h; ++g)
    {
      HighwaySection& gs = sections[g];
      DOUBLE gt, ht; 
      gs.calculateInterchange(hs, gt, ht);
      if (ht <= 0 || ht >= 1 || gt <= 0 || gt >= 1) continue;
      int interchangeId = nodes.size();
      Point interchangeLocation = gs.rampLocation(gt);
      nodes.push_back(HighwayNode(interchangeLocation, nrSites));
      gs.ramps.push_back(ValueId(gt, interchangeId));
      hs.ramps.push_back(ValueId(ht, interchangeId));
    }
  }
  int nrInputNodes = nodes.size();

  // create vertices for all points where a ray from one of the 
  // sites or section endpoints reaches another highway section
  // under angle arccos(1/speed).
  for (int h = 0; h < sections.size(); ++h)
  {
    if (options.verbose)
      reportProgress(((DOUBLE) h)/sections.size(), "> Building highway map");

    HighwaySection& hw = sections[h];
    for (int v = 0; v < nrPrimaryNodes; ++v)
    {
      if (h == v/2) continue; // v is one of the endpoints of hw
      DOUBLE t[2];
      hw.calculateRamps(t, nodes[v].location);
      for (int i = 0; i < 2; ++i)
      {
        if (t[i] <= 0 || t[i] >= 1) continue;
        Point rampLocation = hw.rampLocation(t[i]);
        DOUBLE distance = (rampLocation - nodes[v].location).r();
        int rampNodeId = nodes.size();
        nodes.push_back(HighwayNode(rampLocation, nrSites));
        nodes[v].connect(rampNodeId, distance); ++nrEdges;
        nodes[rampNodeId].connect(v, distance); ++nrEdges;
        hw.ramps.push_back(ValueId(t[i], rampNodeId));     
      }
    }
    sort(hw.ramps.begin(), hw.ramps.end()); // by t-value (first of Ramp)
    for (int i = 1; i < hw.ramps.size(); ++i)
    {
      DOUBLE distance = 
        (nodes[hw.ramps[i].id].location - nodes[hw.ramps[i-1].id].location).r()
        / hw.speed;
      nodes[hw.ramps[i-1].id].connect(hw.ramps[i].id, distance); ++nrEdges;
      nodes[hw.ramps[i].id].connect(hw.ramps[i-1].id, distance); ++nrEdges;
    }
  }
  if (options.verbose)
    reportProgress(1, "> Building highway map");

  // use Dijkstra's algorithm to compute distances from each node to each site
  // (not implemented: under some circumstances, this could be sped up by 
  // stopping the algorithm as soon as the closest two sites to each node are 
  // known---but not with farthest-site and/or divisively weighted VD)

  nrEdges *= nrSites;
  long nrPopped = 0;
  while (!queue.empty())
  {
    if (options.verbose && (nrPopped & 0xFFFF) == 0)
      reportProgress(((DOUBLE) nrPopped)/nrEdges, "> Computing shortest paths");
    Probe p = queue.top(); queue.pop(); ++nrPopped;
    HighwayNode& n = nodes[p.nodeId];
    if (p.distance >= n.distances[p.siteId]) continue;
    n.distances[p.siteId] = p.distance;
    for (int i = 0; i < n.neighbours.size(); ++i)
      queue.push(
        Probe(
          p.distance + n.neighbours[i].value, n.neighbours[i].id, p.siteId 
        )
      );
  }
  if (options.verbose)
  {
    reportProgress(1, "> Computing shortest paths");
//    cerr << "Preparing to compute best access points..."; cerr.flush();
  }

  cells = newMatrix<Cell>(nrSites, gridWidth, gridWidth);
  int totalRegisteredNodes = 0;
  int totalRegisteredSections = 0;
  const DOUBLE cellHalfWidth = 
    options.windowWidth / options.width * cellWidth / 2;
  const DOUBLE cellHalfDiameter = // with error margin:
    options.windowWidth / options.width * (cellWidth+3) * sqrt(0.5);
  DOUBLE* nodeDistance[nrSites];
  for (int i = 0; i < nrSites; ++i) 
    nodeDistance[i] = new DOUBLE[nrPrimaryNodes];
  DOUBLE* sectionDistance[nrSites];
  for (int i = 0; i < nrSites; ++i) 
    sectionDistance[i] = new DOUBLE[sections.size()];
  DOUBLE upperBound[nrSites];

  for (int gridRow = 0; gridRow < gridWidth; ++gridRow)
  {
    if (options.verbose)
      reportProgress(((DOUBLE) gridRow)/gridWidth, "> Computing best access points");

    for (int gridCol = 0; gridCol < gridWidth; ++gridCol)
    {
      Point cellCentre(
        (2.0*gridCol+1.0) * cellHalfWidth + options.windowOffset, 
        (2.0*gridRow+1.0) * cellHalfWidth + options.windowOffset
      );
      for (int siteId = 0; siteId < nrSites; ++siteId)
      {
        upperBound[siteId] = INF;
        for (int i = 0; i < nrPrimaryNodes; ++i)
        {
          nodeDistance[siteId][i] = distance(cellCentre, nodes[i], siteId);
          keepMin(upperBound[siteId], nodeDistance[siteId][i]);
        }
        for (int h = 0; h < sections.size(); ++h)
        {
          sectionDistance[siteId][h] = distance(cellCentre, sections[h], siteId);
          keepMin(upperBound[siteId], sectionDistance[siteId][h]);
        }
        upperBound[siteId] += cellHalfDiameter;
        sites[siteId].weighDistance(upperBound[siteId]);
      }
      DOUBLE firstUpperBound = INF;
      DOUBLE secondUpperBound = INF;
      for (int siteId = 0; siteId < nrSites; ++siteId)
      {
        keepMin(secondUpperBound, upperBound[siteId]);
        if (secondUpperBound < firstUpperBound) 
            swap(firstUpperBound, secondUpperBound);
      }
      DOUBLE mapUpperBound = INF;

      if (!options.useCompositeDistance)
      {
        switch (options.order) 
        {
          case Options::first:    
            mapUpperBound = firstUpperBound + options.equalityThreshold;
            break;
          case Options::second:
          case Options::nextclosest:
            mapUpperBound = secondUpperBound + options.equalityThreshold;
            break;
          case Options::farthest:
            mapUpperBound = INF;
            break;
          default: 
            Complaint(internalError) << 
              "unknown diagram order in HighwayMap constructor";
        }
      }

      for (int siteId = 0; siteId < nrSites; ++siteId)
      {
        Cell& cell = cells[siteId][gridRow][gridCol];
        DOUBLE cellUpperBound = min(mapUpperBound, upperBound[siteId]);

        // distance(cellCentre, s) + cellDiameter/2 is an upper bound
        // for routes via n, 
        //   distance from centre via n - cellDiameter/2 is a lower bound,
        //   so need to be considered only if 
        //   distance from centre via n - cellDiameter/2 < 
        //     distance(cellCentre, s) + cellDiameter/2         <=>

        for (int i = 0; i < nrPrimaryNodes; ++i)
        {
          DOUBLE lowerBound = max((DOUBLE) 0, nodeDistance[siteId][i] - cellHalfDiameter);
          sites[siteId].weighDistance(lowerBound);
          if (lowerBound > cellUpperBound) continue;
          cell.nodes.push_back(i);
          ++totalRegisteredNodes;
        }

        // distance(cellCentre, s) + cellDiameter/2 is an upper bound
        // for routes via h, 
        //   distance from centre via h - speed*cellDiameter/2 is a lower bound,
        //   so need to be considered only if 
        //   distance from centre via h - speed*cellDiameter/2 < 
        //     distance(cellCentre, s) + cellDiameter/2

        for (int h = 0; h < sections.size(); ++h)
        {
          DOUBLE lowerBound = max((DOUBLE) 0,
            sectionDistance[siteId][h] - cellHalfDiameter * sections[h].speed);
          sites[siteId].weighDistance(lowerBound);
          if (lowerBound > cellUpperBound) continue;
          cell.sections.push_back(h);
          ++totalRegisteredSections;
        }
      }
    }
  }
  if (options.verbose)
    reportProgress(1, "> Computing best access points");

  for (int i = 0; i < nrSites; ++i) delete[] nodeDistance[i];
  for (int i = 0; i < nrSites; ++i) delete[] sectionDistance[i];

/*
  if (options.verbose)
  {
    cerr << "\r"
         << "Highway map grid size: " 
         << gridWidth << " x " << gridWidth << " cells" << endl
         << "Average number of registered nodes per site per cell: "
         << (DOUBLE) totalRegisteredNodes / (nrSites * gridWidth * gridWidth) << endl
         << "Average number of registered highway sections per site per cell: "
         << (DOUBLE) totalRegisteredSections / (nrSites * gridWidth * gridWidth) << endl
         << "Preprocessing input..."; 
    cerr.flush();
  }
*/
}


DOUBLE HighwayMap::distance(Point& p, const int siteId)
{
  int gridRow = (p.y() - options.windowOffset) * cellsPerUnit; 
  if (gridRow < 0) gridRow = 0; else 
  if (gridRow >= gridWidth) gridRow = gridWidth-1;
  int gridCol = (p.x() - options.windowOffset) * cellsPerUnit;
  if (gridCol < 0) gridCol = 0; else 
  if (gridCol >= gridWidth) gridCol = gridWidth-1;

  Cell& cell = cells[siteId][gridRow][gridCol];
  DOUBLE d = INF;

  for (
    std::vector<int>::iterator i = cell.nodes.begin(); 
    i != cell.nodes.end(); 
    ++i
  )
    keepMin(d, distance(p, nodes[*i], siteId));

  for (
    std::vector<int>::iterator i = cell.sections.begin(); 
    i != cell.sections.end(); 
    ++i
  )
    keepMin(d, distance(p, sections[*i], siteId));

  return d;
}


void HighwayMap::showNodes(Canvas& canvas)
{
  for (int row = 1; row < gridWidth; ++row)
  {
    Point p(options.windowWidth * row / gridWidth + options.windowOffset, 
        options.windowOffset);
    Point q (options.windowWidth * row / gridWidth + options.windowOffset, 
        options.windowOffset + options.windowWidth);
    canvas.draw(p, q, contourPen, RGB::white);
  }
  for (int col = 1; col < gridWidth; ++col)
  {
    Point p(options.windowOffset, 
      options.windowWidth * col / gridWidth + options.windowOffset);
    Point q(options.windowOffset + options.windowWidth, 
        options.windowWidth * col / gridWidth + options.windowOffset);
    canvas.draw(p, q, contourPen, RGB::white);
  }
  for (int i = 0; i < nodes.size(); ++i)
  {
    Point pixel(
      (nodes[i].location.x() - options.windowOffset) * options.pixelsPerUnit + 0.5,
      (nodes[i].location.y() - options.windowOffset) * options.pixelsPerUnit + 0.5
    );
    canvas.draw(pixel.y(), pixel.x(), bisectorPen, RGB::black);
  }
}


HighwayMap* highwayMap = NULL;


/*********************************************************
* Distance                                               *
*********************************************************/

DOUBLE sqr(const DOUBLE x)
{
  return x * x;
}


DOUBLE safeAcos(const DOUBLE x)
{
  if (x >= 1) return 0;
  if (x <= -1) return PI;
  return acos(x);
}


DOUBLE& keepMin(DOUBLE& x, const DOUBLE y)
{
  if (y < x) x = y;
  return x;
}


DOUBLE& keepMax(DOUBLE& x, const DOUBLE y)
{
  if (y > x) x = y;
  return x;
}


void getSumRange(DOUBLE& alw, DOUBLE& aup, const DOUBLE blw, const DOUBLE bup)
// !! assumes none of alw, aup, blw, bup are NaN !!
{
  alw += blw;
  aup += bup;

  // now if, for example, alw was INF and blw is -INF, aup must have been
  // INF, and alw is now NaN. Two cases can be distinguished:
  // 1. aup is now NaN. This can only happen if bup = -INF. So there is no
  //    valid value for the sum over the whole range. This is an error.
  // 2. aup is now INF. This can only happen if bup > -INF. So the only 
  //    valid value for the sum is INF, we should also set alw to INF.
  // so: if both alw and aup are NaN now, we should complain;
  // if only one of them is NaN, it should take the value of the other:

  if (isnan(alw)) alw = aup;
  if (isnan(aup)) aup = alw;

  if (isnan(alw)) Complaint(userError) 
    << "area of positive size with undefined distance detected";
}


const string Distance::description() const
{
  Complaint(internalError) 
    << "virtual function Distance::description() called";
  return "abstract";
}


SiteTypes Distance::supportedSites() const
{
  Complaint(internalError) 
    << "virtual function Distance::supportedSites() called";
  return 0;
}


bool Distance::supports(const SiteTypes s) const
{
  return (~supportedSites() & s) == 0;
}


DOUBLE Distance::operator()(Point& p, const Site& s) const
{
  Complaint(internalError) 
    << "virtual function Distance::operator()() called";
  return NOTYPES;
}


void Distance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = 0; 
  upperBound = INF;
}


void Distance::verifySite(Site& s) const
{
  if (!supports(POLYLINETYPE) && s.size > 1) Complaint(userError) 
    << "Input contains polyline site, "
    << "for which the " << description() << " distance "
    << "is undefined or has not been implemented";

  if (!supports(BASICPOINTTYPE) && !supports(ROOTEDVECTORTYPE) && s.size == 1) 
    Complaint(userError) 
    << "Input contains single-point site, "
    << "for which the " << description() << " distance " 
    << "is meaningless";
}


SiteTypes FieldDistance::supportedSites() const
{
  return ALLTYPES;
}


ConstantDistance::ConstantDistance(const DOUBLE _weight):
  weight(_weight)
{};


const string ConstantDistance::description() const
{
  stringstream s;
  s << weight;
  return s.str();
}


DOUBLE ConstantDistance::operator()(Point& p, const Site& s) const
{
  return weight;
}


void ConstantDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = weight;
  upperBound = weight;
}


const string FieldXDistance::description() const
{
  return "fx";
}


DOUBLE FieldXDistance::operator()(Point& p, const Site& s) const
{
  return p.x();
}


void FieldXDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = centre.x() - halfwidth;
  upperBound = centre.x() + halfwidth;
}


const string FieldYDistance::description() const
{
  return "fy";
}


DOUBLE FieldYDistance::operator()(Point& p, const Site& s) const
{
  return p.y();
}


void FieldYDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = centre.y() - halfwidth;
  upperBound = centre.y() + halfwidth;
}


const string FieldRDistance::description() const
{
  return "fr";
}


DOUBLE FieldRDistance::operator()(Point& p, const Site& s) const
{
  return p.r();
}


void FieldRDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  bool onYAxis = (centre.x() - halfwidth <= 0 && 0 <= centre.x() + halfwidth);
  bool onXAxis = (centre.y() - halfwidth <= 0 && 0 <= centre.y() + halfwidth);
  if (onXAxis && onYAxis)
    lowerBound = 0;
  else if (onXAxis)
    lowerBound = abs(centre.x()) - halfwidth;
  else if (onYAxis)
    lowerBound = abs(centre.y()) - halfwidth;
  else 
    lowerBound = sqrt(
      sqr(abs(centre.x()) - halfwidth) +
      sqr(abs(centre.y()) - halfwidth)
    );
  upperBound = sqrt(
    sqr(abs(centre.x()) + halfwidth) +
    sqr(abs(centre.y()) + halfwidth)
  );
}


void SiteDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = operator()(centre, s);
  upperBound = lowerBound;
}
  

const string SiteXDistance::description() const
{
  return "sx";
}


SiteTypes SiteXDistance::supportedSites() const
{
  return POINTTYPE;
}


DOUBLE SiteXDistance::operator()(Point& p, const Site& s) const
{
  return s.points[0].x();
}


const string SiteYDistance::description() const
{
  return "sy";
}


SiteTypes SiteYDistance::supportedSites() const
{
  return POINTTYPE;
}


DOUBLE SiteYDistance::operator()(Point& p, const Site& s) const
{
  return s.points[0].y();
}


const string SiteRDistance::description() const
{
  return "sr";
}


SiteTypes SiteRDistance::supportedSites() const
{
  return POINTTYPE;
}


DOUBLE SiteRDistance::operator()(Point& p, const Site& s) const
{
  return s.points[0].r();
}


template < char S >
BinaryCompositionDistance< S >::BinaryCompositionDistance(Distance* _a, Distance* _b):
  a(_a), b(_b)
{
  options.useCompositeDistance = true;
}


template < char S >
const string BinaryCompositionDistance< S >::description() const
{
  stringstream s;
  s << '(' << a->description() << ' ' << S << ' ' << b->description() << ')';
  return s.str();
}


template < char S >
SiteTypes BinaryCompositionDistance< S >::supportedSites() const
{
  return a->supportedSites() & b->supportedSites();
}


template < char S >
void BinaryCompositionDistance< S >::verifySite(Site& s) const
{
  a->verifySite(s);
  b->verifySite(s);
}


DOUBLE MinDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE valueA = (*a)(p, s);
  DOUBLE valueB = (*b)(p, s);
  if (isnan(valueA)) return valueB;
  if (isnan(valueB)) return valueA;
  return min(valueA, valueB);
}


void MinDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  lowerBound = min(alw, blw);
  upperBound = min(aup, bup);
}


DOUBLE MaxDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE valueA = (*a)(p, s);
  DOUBLE valueB = (*b)(p, s);
  if (isnan(valueA)) return valueB;
  if (isnan(valueB)) return valueA;
  return max(valueA, valueB);
}


void MaxDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  lowerBound = max(alw, blw);
  upperBound = max(aup, bup);
}


DOUBLE SumDistance::operator()(Point& p, const Site& s) const
{
  return (*a)(p, s) + (*b)(p, s);
}


void SumDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  a->getRange(centre, halfwidth, s, lowerBound, upperBound);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  getSumRange(lowerBound, upperBound, blw, bup);
}


DOUBLE DifferenceDistance::operator()(Point& p, const Site& s) const
{
  return (*a)(p, s) - (*b)(p, s);
}


void DifferenceDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  a->getRange(centre, halfwidth, s, lowerBound, upperBound);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  getSumRange(lowerBound, upperBound, -bup, -blw);
}


DOUBLE AbsoluteDifferenceDistance::operator()(Point& p, const Site& s) const
{
  return abs((*a)(p, s) - (*b)(p, s));
}


void AbsoluteDifferenceDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE difflw; DOUBLE diffup;
  a->getRange(centre, halfwidth, s, difflw, diffup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  getSumRange(difflw, diffup, -bup, -blw);
  if (difflw < 0 && diffup > 0) 
    lowerBound = 0;
  else
    lowerBound = max(-diffup, difflw);
  upperBound = max(-difflw, diffup);
}


DOUBLE ProductDistance::operator()(Point& p, const Site& s) const
{
  return (*a)(p, s) * (*b)(p, s);
}


void ProductDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  lowerBound =  INF;
  upperBound = -INF; 

  // if one of the ranges is [0,0] and the other is (-inf,inf), then
  // the product can be 0 (or invalid, nothing else):
  if (
    (alw == 0 && aup == 0 && blw == -INF && bup == INF) || 
    (blw == 0 && bup == 0 && alw == -INF && aup == INF)
  )
  {
    lowerBound = 0;
    upperBound = 0;
    return;
  }

  // otherwise extremes are obtained by multiplying an extreme of one 
  // range with an extreme of the other; whether these are lower or 
  // upper bounds depends on the signs
  DOUBLE x;
  x = alw * blw; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = alw * bup; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = aup * blw; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = aup * bup; keepMin(lowerBound, x); keepMax(upperBound, x);
}


DOUBLE RatioDistance::operator()(Point& p, const Site& s) const
{
  return (*a)(p, s) / (*b)(p, s);
}


void RatioDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);
  lowerBound =  INF;
  upperBound = -INF; 

  // consider the result of division by zero
  if (blw <= 0 && bup >= 0)
  {
    if (alw < 0) lowerBound = -INF;
    if (aup > 0) upperBound = INF;
  }

  // consider the result of division by zero approached from below
  if (blw < 0 && bup >= 0)
  {
    if (aup > 0) lowerBound = -INF;
    if (alw < 0) upperBound = INF;
  }

  // consider the result of dividing by arbitrarily large numbers
  if (alw < INF && aup > -INF && (blw == -INF || bup == INF))
  {
    keepMin(lowerBound, (DOUBLE) 0);
    keepMax(upperBound, (DOUBLE) 0);
  }

  // otherwise extremes are obtained by dividing an extreme of one 
  // range by an extreme of the other; whether these are lower or 
  // upper bounds depends on the signs:
  DOUBLE x;
  x = alw / blw; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = alw / bup; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = aup / blw; keepMin(lowerBound, x); keepMax(upperBound, x);
  x = aup / bup; keepMin(lowerBound, x); keepMax(upperBound, x);
}


DOUBLE PowerDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE da = (*a)(p, s);
  if (da < 0) Complaint(userError) 
    << "negative value " << da << " for first operand to power operator detected";
  return pow(da, (*b)(p, s));
}


void PowerDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE blw; DOUBLE bup; b->getRange(centre, halfwidth, s, blw, bup);

  // the first operand is not allowed to be negative:
  if (aup < 0) Complaint(userError) 
    << "negative value " << aup << " for upper bound on first operand to power operator detected";
  if (alw < 0) alw = 0;

  lowerBound = INF;
  upperBound = -INF;

  // extremes are obtained by taking an extreme of the base range 
  // to an extreme of the power range; whether these are lower or 
  // upper bounds depends on where the base lies w.r.t. 1:
  DOUBLE x;
  x = pow(alw, blw); keepMin(lowerBound, x); keepMax(upperBound, x);
  x = pow(alw, bup); keepMin(lowerBound, x); keepMax(upperBound, x);
  x = pow(aup, blw); keepMin(lowerBound, x); keepMax(upperBound, x);
  x = pow(aup, bup); keepMin(lowerBound, x); keepMax(upperBound, x);
}


UnaryOperatorDistance::UnaryOperatorDistance(Distance* _a, const char* _symbol):
  a(_a), symbol(_symbol)
{
  options.useCompositeDistance = true;
}


const string UnaryOperatorDistance::description() const
{
  stringstream s;
  s << symbol << '(' << a->description() << ')';
  return s.str();
}


SiteTypes UnaryOperatorDistance::supportedSites() const
{
  return a->supportedSites();
}


void UnaryOperatorDistance::verifySite(Site& s) const
{
  a->verifySite(s);
}


LogDistance::LogDistance(Distance* _a):
  UnaryOperatorDistance(_a, "ln")
{};


DOUBLE LogDistance::operator()(Point& p, const Site& s) const
{
  return log((*a)(p, s));
}


void LogDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  lowerBound = log(alw);
  upperBound = log(aup);
}


AbsDistance::AbsDistance(Distance* _a):
  UnaryOperatorDistance(_a, "abs")
{};


DOUBLE AbsDistance::operator()(Point& p, const Site& s) const
{
  return abs((*a)(p, s));
}


void AbsDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; a->getRange(centre, halfwidth, s, alw, aup);
  if (alw < 0 && aup > 0)
    lowerBound = 0;
  else
    lowerBound = max(-aup, alw);
  upperBound = max(-alw, aup);
}


SquareDistance::SquareDistance(Distance* _a):
  UnaryOperatorDistance(_a, "sq")
{};


DOUBLE SquareDistance::operator()(Point& p, const Site& s) const
{
  return sqr((*a)(p, s));
}


void SquareDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; 
  a->getRange(centre, halfwidth, s, alw, aup);
  DOUBLE sqalw = sqr(alw);
  DOUBLE sqaup = sqr(aup);
  if (alw < 0 && aup > 0) 
    lowerBound = 0;
  else
    lowerBound = min(sqalw, sqaup);
  upperBound = max(sqalw, sqaup);
}


SquareRootDistance::SquareRootDistance(Distance* _a):
  UnaryOperatorDistance(_a, "sqrt")
{};


DOUBLE SquareRootDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE da = (*a)(p, s);
  if (da < 0) Complaint(userError) 
    << "negative value " << da << " for operand to square root operator detected";
  return sqrt(da);
}


void SquareRootDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE alw; DOUBLE aup; 
  a->getRange(centre, halfwidth, s, alw, aup);

  // the operand is not allowed to be negative:
  if (aup < 0) Complaint(userError) 
    << "negative value " << aup << " for upper bound on operand to square-root operator detected";
  if (alw < 0) alw = 0;

  lowerBound = sqrt(alw);
  upperBound = sqrt(aup);
}


SiteTypes PointDistance::supportedSites() const
{
  return POINTTYPE;
}

   
SiteTypes PolylineDistance::supportedSites() const
{
  return POLYLINETYPE;
}


SiteTypes SetDistance::supportedSites() const
{
  return POINTTYPE | POLYLINETYPE;
}

   
SiteTypes RootedVectorDistance::supportedSites() const
{
  return ROOTEDVECTORTYPE;
}


const string EuclideanDistance::description() const
{
  return "Euclidean";
}


DOUBLE EuclideanDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE sqDistance = v * v;
  for (int i = 1; i < s.size; ++i) 
  {
    // calculate x-coordinate if axis system is rotated
    // and translated such that the segment points to the right,
    // and scaled with factor s.vectors[i].r():
    DOUBLE x = s.vectors[i] * v;
    // check if point lies vertically above or below the segment
    if (x > 0 && x < sqr(s.vectors[i].r()))
    {
      // calculate y-coordinate under same transformation
      DOUBLE y = s.vectors[i].rotLeft() * v;
      keepMin(sqDistance, sqr(y / s.vectors[i].r()));
    }
    v = p - s.points[i];
    keepMin(sqDistance, v * v);
  }
  return sqrt(sqDistance);
}


void EuclideanDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE distance = (*this)(centre, s);
  DOUBLE radius = halfwidth * sqrt(2.0);
  lowerBound = max(distance - radius, (DOUBLE) 0);
  upperBound = distance + radius;
}


const string SquaredEuclideanDistance::description() const
{
  return "squared Euclidean";
}


DOUBLE SquaredEuclideanDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE sqDistance = v * v;
  for (int i = 1; i < s.size; ++i) 
  {
    // calculate x-coordinate if axis system is rotated
    // and translated such that the segment points to the right,
    // and scaled with factor s.vectors[i].r():
    DOUBLE x = s.vectors[i] * v;
    // check if point lies vertically above or below the segment
    if (x > 0 && x < sqr(s.vectors[i].r()))
    {
      // calculate y-coordinate under same transformation
      DOUBLE y = s.vectors[i].rotLeft() * v;
      keepMin(sqDistance, sqr(y / s.vectors[i].r()));
    }
    v = p - s.points[i];
    keepMin(sqDistance, v * v);
  }
  return sqDistance;
}


void SquaredEuclideanDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE distance = sqrt((*this)(centre, s));
  DOUBLE radius = halfwidth * sqrt(2.0);
  lowerBound = sqr(max(distance - radius, (DOUBLE) 0));
  upperBound = sqr(distance + radius);
}


HighwayDistance::HighwayDistance()
{
  options.useHighways = true;
}


const string HighwayDistance::description() const
{
  return "highway";
}


DOUBLE HighwayDistance::operator()(Point& p, const Site& s) const
{
  return highwayMap->distance(p, s.id);
}


void HighwayDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE distance = (*this)(centre, s);
  DOUBLE radius = halfwidth * sqrt(2.0);
  lowerBound = max(distance - radius, (DOUBLE) 0);
  upperBound = distance + radius;
}


const string ManhattanDistance::description() const
{
  return "L1";
}


DOUBLE ManhattanDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE distance = abs(v.x()) + abs(v.y());
  for (int i = 1; i < s.size; ++i)  
  {
    if (abs(s.vectors[i].x()) > abs(s.vectors[i].y()))
    {
      // test if p.x lies between s.points[i-1].x and s.points[i].x,
      // works regardless of which endpoint has the smaller x-coordinate
      if ((v.x() > 0) == (v.x() < s.vectors[i].x()))
        keepMin(distance, abs(v.y() - v.x() * s.vectors[i].ySlope()));
    }
    else
    {
      if ((v.y() > 0) == (v.y() < s.vectors[i].y()))
        keepMin(distance, abs(v.x() - v.y() * s.vectors[i].xSlope()));
    }
    v = p - s.points[i];
    keepMin(distance, abs(v.x()) + abs(v.y()));
  }
  return distance;
}


void ManhattanDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  DOUBLE distance = (*this)(centre, s);
  lowerBound = max(distance - halfwidth * 2, (DOUBLE) 0);
  upperBound = distance + halfwidth * 2;
}


const string LMinDistance::description() const
{
  return "Lminf";
}


DOUBLE LMinDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return min(abs(v.x()), abs(v.y()));
}


void LMinDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  Point v = centre - s.points[0];

  DOUBLE xLowerBound = 0;
  keepMax(xLowerBound,  v.x() - halfwidth);
  keepMax(xLowerBound, -v.x() - halfwidth);
  DOUBLE yLowerBound = 0;
  keepMax(yLowerBound,  v.y() - halfwidth);
  keepMax(yLowerBound, -v.y() - halfwidth);
  lowerBound = min(xLowerBound, yLowerBound);

  DOUBLE xUpperBound = 0;
  keepMax(xUpperBound,  v.x() + halfwidth);
  keepMax(xUpperBound, -v.x() + halfwidth);
  DOUBLE yUpperBound = 0;
  keepMax(yUpperBound,  v.y() + halfwidth);
  keepMax(yUpperBound, -v.y() + halfwidth);
  upperBound = min(xUpperBound, yUpperBound);
}


const string VolumeDistance::description() const
{
  return "bbox volume";
}


DOUBLE VolumeDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return abs(v.x()*v.y());
}


void VolumeDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  Point v = centre - s.points[0];

  DOUBLE xLowerBound = 0;
  keepMax(xLowerBound,  v.x() - halfwidth);
  keepMax(xLowerBound, -v.x() - halfwidth);
  DOUBLE yLowerBound = 0;
  keepMax(yLowerBound,  v.y() - halfwidth);
  keepMax(yLowerBound, -v.y() - halfwidth);
  lowerBound = xLowerBound * yLowerBound;

  DOUBLE xUpperBound = 0;
  keepMax(xUpperBound,  v.x() + halfwidth);
  keepMax(xUpperBound, -v.x() + halfwidth);
  DOUBLE yUpperBound = 0;
  keepMax(yUpperBound,  v.y() + halfwidth);
  keepMax(yUpperBound, -v.y() + halfwidth);
  upperBound = xUpperBound * yUpperBound;
}


const string LMaxDistance::description() const
{
  return "Linf";
}


DOUBLE LMaxDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return max(abs(v.x()), abs(v.y()));
}


void LMaxDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  Point v = centre - s.points[0];

  DOUBLE xLowerBound = 0;
  keepMax(xLowerBound,  v.x() - halfwidth);
  keepMax(xLowerBound, -v.x() - halfwidth);
  DOUBLE yLowerBound = 0;
  keepMax(yLowerBound,  v.y() - halfwidth);
  keepMax(yLowerBound, -v.y() - halfwidth);
  lowerBound = max(xLowerBound, yLowerBound);

  DOUBLE xUpperBound = 0;
  keepMax(xUpperBound,  v.x() + halfwidth);
  keepMax(xUpperBound, -v.x() + halfwidth);
  DOUBLE yUpperBound = 0;
  keepMax(yUpperBound,  v.y() + halfwidth);
  keepMax(yUpperBound, -v.y() + halfwidth);
  upperBound = max(xUpperBound, yUpperBound);
}


LDistance::LDistance(const DOUBLE _n):
  n(_n)
{
  if (n == 0) Complaint(internalError) << "L0-distance is undefined";
  invn = ((DOUBLE) 1.0) / n;
}


const string LDistance::description() const
{
  stringstream s;
  s << "L" << n;
  return s.str();
}


DOUBLE LDistance::operator()(Point& p, const Site& s) const
{
  if (n < 0 && (p.x() == s.points[0].x() || p.y() == s.points[0].y())) 
    return 0;
  return pow(
    pow(abs(p.x() - s.points[0].x()), n) +
    pow(abs(p.y() - s.points[0].y()), n),
  invn);
}


void LDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  Point v = centre - s.points[0];

  DOUBLE xLowerBound = 0;
  keepMax(xLowerBound,  v.x() - halfwidth);
  keepMax(xLowerBound, -v.x() - halfwidth);
  DOUBLE yLowerBound = 0;
  keepMax(yLowerBound,  v.y() - halfwidth);
  keepMax(yLowerBound, -v.y() - halfwidth);
  lowerBound = 0;
  if (n > 0 || (xLowerBound != 0 && yLowerBound != 0)) 
    lowerBound = pow(pow(xLowerBound, n) + pow(yLowerBound, n), invn);

  DOUBLE xUpperBound = 0;
  keepMax(xUpperBound,  v.x() + halfwidth);
  keepMax(xUpperBound, -v.x() + halfwidth);
  DOUBLE yUpperBound = 0;
  keepMax(yUpperBound,  v.y() + halfwidth);
  keepMax(yUpperBound, -v.y() + halfwidth);
  upperBound = pow(pow(xUpperBound, n) + pow(yUpperBound, n), invn);
}


const string KarlsruheDistance::description() const
{
  return "Karlsruhe";
}


DOUBLE KarlsruheDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE angleDistance = safeAcos((p * s.points[0]) / (p.r() * s.points[0].r()));
  if (angleDistance > 2) return p.r() + s.points[0].r();
  return 
    angleDistance * min(p.r(), s.points[0].r()) + 
    abs(p.r() - s.points[0].r());
}


void KarlsruheDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  // use that KA/Euclidean distance <= sqrt(2)
  DOUBLE distance = (*this)(centre, s);
  lowerBound = max(distance - halfwidth * 2, (DOUBLE) 0);
  upperBound = distance + halfwidth * 2;
}


const string AzimuthDistance::description() const
{
  return "azimuth difference";
}


DOUBLE AzimuthDistance::operator()(Point& p, const Site& s) const
{
  if (p.r() == 0 || s.points[0].r() == 0) return 0;
  return safeAcos((p * s.points[0]) / (p.r() * s.points[0].r()));
}


void AzimuthDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  Point origin(0,0);
  getTurnRange(centre, halfwidth, origin, s.points[0], lowerBound, upperBound);
}


const string AngleDistance::description() const
{
  return "angular size";
}


DOUBLE AngleDistance::operator()(Point& p, const Site& s) const
{
  if (s.points[0] == p) return (DOUBLE) 1.0 / TWOPI;
  DOUBLE angle = (s.points[0] - p).fi();
  DOUBLE minAngle = angle;
  DOUBLE maxAngle = angle;
  for (int i = 1; i < s.size; ++i)
  {
    if (s.points[i] == p) return (DOUBLE) 1.0 / TWOPI;
    DOUBLE prevAngle = angle;
    angle = (s.points[i] - p).fi();
    DOUBLE dAngle = fmod(angle - prevAngle + TWOPI * i, TWOPI);
    if (dAngle < PI) 
    {
      angle = prevAngle + dAngle;
      keepMax(maxAngle, angle);
    }      
    else // dAngle >= PI
    {
      angle = prevAngle + (dAngle - TWOPI);
      keepMin(minAngle, angle);
    }
  }
  // if (maxAngle == minAngle) return INF; // implied by what's below
  if (maxAngle - minAngle > TWOPI) return (DOUBLE) 1.0 / TWOPI;
  return (DOUBLE) 1.0 / (maxAngle - minAngle);
}


const string DetourDistance::description() const
{
  return "detour";
}


DOUBLE DetourDistance::operator()(Point& p, const Site& s) const
{
  if (s.size == 1) return (s.points[0] - p).r() * 2.0;
  DOUBLE distance = INF;
  for (int i = 1; i < s.size; ++i)
    keepMin(distance, 
      (s.points[i-1] - p).r() + (p - s.points[i]).r() - s.vectors[i].r()
    );
  return distance;
}


const string DilationDistance::description() const
{
  return "dilation";
}


DOUBLE DilationDistance::operator()(Point& p, const Site& s) const
{
  DOUBLE distance = INF;
  for (int i = 1; i < s.size; ++i)
    keepMin(distance, 
      ((s.points[i-1] - p).r() + (p - s.points[i]).r()) / s.vectors[i].r() - 1.0
    );
  return distance;
}


const string SemiDistance::description() const
{
  return "halfplane indicator";
}


DOUBLE SemiDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE dotproduct = s.vectors[0] * v;
  return (dotproduct < 0 ? INF : 1);
}


void SemiDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  lowerBound = INF;
  upperBound = 1;
  Point corner[4] = {
    centre + Point(-halfwidth, -halfwidth),
    centre + Point(-halfwidth,  halfwidth),
    centre + Point( halfwidth, -halfwidth),
    centre + Point( halfwidth,  halfwidth)
  };
  for (int i = 0; i < 4; ++i)
  {
    DOUBLE cornerDistance = (*this)(corner[i], s);
    keepMin(lowerBound, cornerDistance);
    keepMax(upperBound, cornerDistance);
  }
}


const string SecantDistance::description() const
{
  return "secant";
}


DOUBLE SecantDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE dotproduct = s.vectors[0] * v;
  if (dotproduct <= 0) return INF;
  return v.r() * s.vectors[0].r() / dotproduct;
}


void SecantDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  getTurnRange(centre, halfwidth, s.points[0], s.vectors[0], lowerBound, upperBound);
  lowerBound = (lowerBound >= 0.5*PI ? INF : 1.0 / cos(lowerBound));
  upperBound = (upperBound >= 0.5*PI ? INF : 1.0 / cos(upperBound));
}


const string CatchDistance::description() const
{
  return "catch";
}


DOUBLE CatchDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  DOUBLE dotproduct = s.vectors[0] * v;
  if (dotproduct <= 0) return INF;
  return v * v * s.vectors[0].r() * 0.5 / dotproduct;
}


void CatchDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  // catch = secant * Euclid. distance * 0.5
  DOUBLE sLowerBound; DOUBLE sUpperBound;
  SecantDistance().getRange(centre, halfwidth, s, sLowerBound, sUpperBound);
  DOUBLE eLowerBound; DOUBLE eUpperBound;
  EuclideanDistance().getRange(centre, halfwidth, s, eLowerBound, eUpperBound);
  lowerBound = (sLowerBound == INF ? INF : sLowerBound * eLowerBound);
  upperBound = sUpperBound * eUpperBound;
}


MixedCatchDistance::MixedCatchDistance(DOUBLE _speed):
  speed(_speed)
{};


const string MixedCatchDistance::description() const
{
  stringstream s;
  s << "catch@" << speed;
  return s.str();
}


DOUBLE MixedCatchDistance::operator()(Point& p, const Site& s) const
{
  if (p == s.points[0]) return 0;

  // let q, dq, qq be vector, distance, squared distance from s to p
  // let w, dw, ww be vector, length, squared length of s
  // let z be w rotated by 90 degrees
  // let u, uu be unit, squared unit distance
  // let alfa be angle between q and w, so 
  // cosa := cos alfa = q.w / (dq dw)
  // sqsina := (sin alfa)^2 = (q.z)^2 / (qq ww)
  // D := qq * uu - ww * qq * sqsina = qq * uu - (q.z)^2
  // to return:
  // if dw = u, then 
  //   if cosa <= 0 then INF
  //   else 0.5 * dq / cosa = 0.5 * qq * dw / q.w = 0.5 * u * qq / q.w
  // otherwise, 
  //   if D < 0 or if behind a faster site then INF
  //   else dq * (dw/u cosa - sqrt(Dqquu)/(u*dq)) / (ww/uu - 1)
  //      = (q.w - sqrt(D)) / (ww/u - u)

  DOUBLE ww = s.vectors[0] * s.vectors[0];
  DOUBLE uu = sqr(speed);
  Point q = p - s.points[0];
  DOUBLE qq = q * q;
  DOUBLE qw = q * s.vectors[0];
  if (ww >= uu && qw <= 0) return INF;
  if (ww == uu) return qq * speed * 0.5 / qw;
  DOUBLE D = qq * uu - sqr(q * s.vectors[0].rotLeft());
  if (D < 0) return INF;
  return (qw - sqrt(D)) / (ww/speed - speed);
} 


const string TurnDistance::description() const
{
  return "turn";
}


DOUBLE TurnDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  return safeAcos(s.vectors[0] * v / (v.r() * s.vectors[0].r()));
}


void TurnDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  getTurnRange(centre, halfwidth, s.points[0], s.vectors[0], lowerBound, upperBound);
}


const string LeftTurnDistance::description() const
{
  return "left turn";
}


DOUBLE LeftTurnDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];
  if (v.x() == 0 && v.y() == 0) return 0;
  // rotate such that s.vectors[0] points in direction 0
  // (it also scales, but that does not matter)
  Point n(s.vectors[0] * v, s.vectors[0].rotLeft() * v);
  DOUBLE angle = atan2(n.y(), n.x());
  return (angle >= 0 ? angle : angle + TWOPI);
}


void LeftTurnDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  // if the ray cuts the square, then bounds are [0, INF)
  // otherwise the lower and upper bound are attained at corners
  // the ray cuts the square iff lowest and highest values of 
  // corners are at least PI apart
  lowerBound = TWOPI;
  upperBound = 0;
  Point corner[4] = {
    centre + Point(-halfwidth, -halfwidth),
    centre + Point(-halfwidth,  halfwidth),
    centre + Point( halfwidth, -halfwidth),
    centre + Point( halfwidth,  halfwidth)
  };
  for (int i = 0; i < 4; ++i)
  {
    DOUBLE cornerDistance = (*this)(corner[i], s);
    keepMin(lowerBound, cornerDistance);
    keepMax(upperBound, cornerDistance);
  }
  if (upperBound - lowerBound < PI) return;

  // ray cuts the square:
  lowerBound = 0;
  upperBound = TWOPI;   
}


const string DubinsDistance::description() const
{
  return "Dubins";
}


DOUBLE DubinsDistance::operator()(Point& p, const Site& s) const
{
  Point v = p - s.points[0];

  // if p is at the starting point, return 0:
  if (v.x() == 0 && v.y() == 0) return 0;

  // get inverse turning radius:
  DOUBLE invr = ((DOUBLE) 1.0) / s.vectors[0].r();

  // rotate such that s.vectors[0] points in direction 0,
  // then reflect such that destination is not below x-axis
  Point n(
    s.vectors[0] * v * invr, 
    abs(s.vectors[0].rotLeft() * v * invr)
  );

  // the same in radius units:
  Point nu = n * invr;

  // get location relative to turning circle centre
  Point ru = nu - Point(0, 1);
 
  if (ru * ru < 1)
  { // within turning circle
    // get location relative to opposite turning circle centre
    ru = nu - Point(0, -1);
    // get distance to opposite turning circle centre
    DOUBLE ruSqLength = ru * ru;
    DOUBLE ruLength = sqrt(ruSqLength);
    return s.vectors[0].r() * (
      // right turn if destination would be on y-axis:
      safeAcos((DOUBLE) 0.25*ruLength + (DOUBLE) 0.75/ruLength) +
      // further right turn:
      atan2(ru.x(), ru.y()) +
      // left turn:
      TWOPI - safeAcos((DOUBLE) 1.25 - (DOUBLE) 0.25*ruSqLength)
    );
  }
  else
  { // outside turning circle
    DOUBLE ruSqLength = ru * ru;
    DOUBLE ruLength = sqrt(ruSqLength);
    return s.vectors[0].r() * (
      // direction relative to turning circle centre, to be modified such
      atan2(-ru.x(), ru.y()) + HALFPI +
      // additional angle because of approach from tangent point
      HALFPI - safeAcos((DOUBLE) 1.0 / ruLength) +
      // straight part:
      sqrt(ruSqLength - 1)
    );
  }
}


void DubinsDistance::getRange(
  Point& centre, DOUBLE halfwidth, const Site& s,
  DOUBLE& lowerBound, DOUBLE& upperBound
) const
{
  // coarse bounds:
  // Euclidean distance is lower bound on Dubins distance
  // Euclidean distance + full circle is upper bound on Dubins distance
  DOUBLE distance = (centre - s.points[0]).r();
  DOUBLE radius = halfwidth * sqrt(2.0);
  lowerBound = max(distance - radius, (DOUBLE) 0);
  upperBound = distance + radius + TWOPI * s.vectors[0].r();
}


#ifdef INCLUDE_USER_DISTANCES
#include "userdistances.cpp"
#endif


Distance* getDistanceObject(string dname)
{
  // strip outer spaces
  int expStart = 0;
  while (expStart < dname.length() && dname[expStart] == ' ') ++expStart;
  int expEnd = dname.length();
  while (expEnd > 0 && dname[expEnd-1] == ' ') --expEnd;
  if (expEnd <= expStart)
    Complaint(usageError) << "missing distance name";
  dname.erase(expEnd);
  dname.erase(0, expStart);
  
  // strip outer parenthesis, if matched
  if (dname.front() == '(' && dname.back() == ')')
  {
    int parenthesisDepth = 1;
    for (int i = 1; parenthesisDepth > 0 && i < dname.length()-1; ++i)
    {
      switch(dname[i])
      { 
        case ')':
          --parenthesisDepth;
          break;
        case '(': 
          ++parenthesisDepth;
          break;
        default:
          break;
      }
    }
    if (parenthesisDepth > 1)
      Complaint(usageError) << "unmatched parenthesis in distance formula";
    if (parenthesisDepth == 1) // call again without outer parenthesis
      return getDistanceObject(dname.substr(1, dname.length()-2));
  }

  // supported operators in order of decreasing precedence:
  // (parsing need to go in opposite order: lowest first)
  // ln(...)  logarithm 
  // ^  power                         from left to right
  // *  multiplication and / division from left to right
  // ~  absolute difference           from left to right
  // +  addition and - subtraction    from left to right
  // _  minimum and | maximum         from left to right

  // find the lowest-priority operator that is not in parenthesis:
  // tricky part: distinguish operator minus from number minus
  // states:
  //  2: after l received in state 1: expect nr minus, other chars, no operators
  //  1: after start or any operator: expect (, l, nr minus, other chars, no operators
  //  0: (standard case)            : expect anything except a nr minus
  // -1: after )                    : expect ) or operator
  
  int operatorPosition = dname.length();
  int operatorPriority = 6;
  int parenthesisDepth = 0;
  int state = 1;
  for (int i = 0; i < dname.length(); ++i)
  {
    char c = dname[i];
    if (c == ')') 
    {
      if (state > 1)
        Complaint(usageError) << "unexpected ) in distance formula";
      --parenthesisDepth;
      if (parenthesisDepth < 0)
        Complaint(usageError) << "unmatched parenthesis in distance formula";
      state = -1;
      continue;
    }
    if (c == '(') 
    {
      if (state == -1 || state == 2)
        Complaint(usageError) << "unexpected ( in distance formula";
      ++parenthesisDepth;
      state = 1;
      continue;
    }
    switch(c)
    {
      case '_': case '|':
        if (state > 0)
          Complaint(usageError) << "unexpected " << c << " in distance formula";
        state = 1;
        if (parenthesisDepth > 0) continue;
        operatorPosition = i;
        operatorPriority = 1;
        break;
      case '-':
        if (state > 0) continue; // might be the sign of a number
      case '+':
        if (state > 0)
          Complaint(usageError) << "unexpected " << c << " in distance formula";
        state = 1;
        if (parenthesisDepth > 0 || operatorPriority < 2) continue;
        operatorPosition = i;
        operatorPriority = 2;
        break;
      case '~': 
        if (state > 0)
          Complaint(usageError) << "unexpected " << c << " in distance formula";
        state = 1;
        if (parenthesisDepth > 0 || operatorPriority < 3) continue;
        operatorPosition = i;
        operatorPriority = 3;
        break;
      case '*': case '/':
        if (state > 0)
          Complaint(usageError) << "unexpected " << c << " in distance formula";
        state = 1;
        if (parenthesisDepth > 0 || operatorPriority < 4) continue;
        operatorPosition = i;
        operatorPriority = 4;
        break;
      case '^':
        if (state > 0)
          Complaint(usageError) << "unexpected " << c << " in distance formula";
        state = 1;
        if (parenthesisDepth > 0 || operatorPriority < 5) continue;
        operatorPosition = i;
        operatorPriority = 5;
        break;
      case 'l':
        if (state == 1) { state = 2; continue; }
      default:
        if (state == -1) 
          Complaint(usageError) 
            << "read " << c << " while expecting one of _|+-~*/^) in distance formula";
        state = 0;
        break;
    }
  }
  if (parenthesisDepth > 0)
    Complaint(usageError) << "unmatched parenthesis in distance formula";
  if (operatorPosition < dname.length())
  {
    Distance* a = getDistanceObject(dname.substr(0, operatorPosition));
    Distance* b = getDistanceObject(dname.substr(operatorPosition+1));
    switch(dname[operatorPosition])
    {
      case '_': return new MinDistance(a,b);
      case '|': return new MaxDistance(a,b);
      case '+': return new SumDistance(a,b);
      case '-': return new DifferenceDistance(a,b);
      case '~': return new AbsoluteDifferenceDistance(a,b);
      case '*': return new ProductDistance(a,b);
      case '/': return new RatioDistance(a,b);
      case '^': return new PowerDistance(a,b);
      default: 
        Complaint(internalError) << "lost operator while parsing distance formula";
    }
  }
  if (dname == "fx") return new FieldXDistance();
  if (dname == "fy") return new FieldYDistance();
  if (dname == "fr") return new FieldRDistance();
  if (dname == "sx") return new SiteXDistance();
  if (dname == "sy") return new SiteYDistance();
  if (dname == "sr") return new SiteRDistance();
  if (dname == "squared euclidean" || 
      dname == "squared") return new SquaredEuclideanDistance();
  if (dname == "euclidean" || dname == "l2") return new EuclideanDistance();
  if (dname == "highway") return new HighwayDistance();
  if (dname == "manhattan" || 
      dname == "l1") return new ManhattanDistance();
  if (dname == "azimuth difference" ||
      dname == "azimuth") return new AzimuthDistance();
  if (dname == "logradius") return getDistanceObject("abs(ln(fr/sr))");
  if (dname == "karlsruhe") return new KarlsruheDistance(); // azimuth*(fr_sr)+fr~sr_fr+sr
  if (dname == "city") return getDistanceObject("sqrt(sq(azimuth)+sq(ln(fr/sr)))");
  if (dname == "koeln") return getDistanceObject("azimuth + logradius");
  if (dname == "orbitout") return getDistanceObject("0 | 1/sr - 2/(sr+euclidean+fr)");
  if (dname == "orbitin") return getDistanceObject("0 | 1/fr - 2/(sr+euclidean+fr)");
  if (dname == "angular size" ||
      dname == "angle") return new AngleDistance();
  if (dname == "detour") return new DetourDistance();
  if (dname == "dilation") return new DilationDistance();
  if (dname == "halfplane indicator" ||
      dname == "semi") return new SemiDistance();
  if (dname == "secant") return new SecantDistance();
  if (dname.substr(0,6) == "catch@") 
  {
    DOUBLE speed = strtofstop(dname.substr(6));
    if (speed <= 0) Complaint(usageError) << "invalid speed setting for catch distance";
    return new MixedCatchDistance(speed);
  }
  if (dname == "catch") return new CatchDistance();
  if (dname == "turn") return new TurnDistance();
  if (dname == "left turn" ||
      dname == "leftturn") return new LeftTurnDistance();
  if (dname == "dubins") return new DubinsDistance();
  if (dname == "min" || 
      dname == "l-inf" || 
      dname == "lminf") return new LMinDistance();
  if (dname == "bbox volume" ||
      dname == "bbox" ||
      dname == "l0") return new VolumeDistance();
  if (dname == "max" || 
      dname == "linf") return new LMaxDistance();
#ifdef INCLUDE_USER_DISTANCES
  MAKEUSERDISTANCE
#endif
  if (dname.substr(0,2) == "ln") return new LogDistance(getDistanceObject(dname.substr(2)));
  if (dname.substr(0,3) == "abs") return new AbsDistance(getDistanceObject(dname.substr(3)));
  if (dname.substr(0,4) == "sqrt") 
    return new SquareRootDistance(getDistanceObject(dname.substr(4)));
  if (dname.substr(0,2) == "sq") return new SquareDistance(getDistanceObject(dname.substr(2)));
  if (dname[0] == 'l') 
  {
    DOUBLE exponent = strtofstop(dname.substr(1));
    if (exponent == 0) return new VolumeDistance();
    else return new LDistance(exponent);
  }
  stringstream s(dname);
  s >> std::ws;
  if (s.eof()) Complaint(usageError) << "missing distance name";
  DOUBLE weight;
  s >> weight >> std::ws;
  if (s.eof()) // it is indeed just a number
    return new ConstantDistance(weight);
  Complaint(usageError) << "unrecognised distance name \"" << dname << "\"";
  return NULL;
}


/*********************************************************
* Canvas                                                 *
*********************************************************/


Canvas::Canvas(const int _nrRows, const int _nrCols):
  nrRows(_nrRows + strokeWidth[nrPens-1]),
  nrCols(_nrCols + strokeWidth[nrPens-1]),  
  pixels(new RGB[nrRows * nrCols]),
  rows(new RGB*[nrRows])
{
  // let rows[i] point to first pixel in output colum in row i
  rows[0] = pixels + strokeWidth[nrPens-1]/2;
  for (int i = 1; i < nrRows; ++i) rows[i] = rows[i-1] + nrCols;

  // reindex rows so that 0 is the first output row
  rows += strokeWidth[nrPens-1]/2;
   
  // calculate upper bound on total pen size
  int totalPenSize = 0;
  for (int i = 0; i < nrPens; ++i) totalPenSize += strokeWidth[i]*strokeWidth[i];

  // reserve space for pen masks
  int* penMask = new int[totalPenSize];

  // calculate pen masks
  for (int i = 0; i < nrPens; ++i)
  {
    penSize[i] = 0;
    penMasks[i] = penMask;
    int radius = strokeWidth[i]/2;
    DOUBLE sqRadius = sqr(radius);
    for (int row = -radius; row < radius; ++row)
      for (int col = -radius; col < radius; ++col)
        if (sqr(0.5 + (DOUBLE) row) + sqr(0.5 + (DOUBLE) col) <= sqRadius)
        {
          *(penMask++) = row * nrCols + col;
          ++penSize[i];
        }
  }
}
  

const RGB& Canvas::operator()(const int row, const int col) const
{
  return rows[row][col];
}


template < class T >
void Canvas::shade(const int row, const int col, const T& value)
{
  rows[row][col] *= value;
}


void Canvas::draw(const int row, const int col, const int pen, const RGB& colour)
{
  if (row < 0 || row >= options.width || col < 0 || col > options.width) return;
  RGB* centre = rows[row] + col;
  int* penMask = penMasks[pen];
  for (int i = penSize[pen]-1; i >= 0; --i) centre[penMask[i]] = colour;
}


template < class T >
void Canvas::shade(const int row, const int col, const int pen, const T& value)
{
  if (row < 0 || row >= options.width || col < 0 || col > options.width) return;
  RGB* centre = rows[row] + col;
  int* penMask = penMasks[pen];
  for (int i = penSize[pen]-1; i >= 0; --i) centre[penMask[i]] *= value;
}


void Canvas::draw(Point& p, Point& q, const int pen, const RGB& colour)
{
  // calculate pixel coordinates
  Point pp(
    (p.x() - options.windowOffset) * options.pixelsPerUnit + 0.5,
    (p.y() - options.windowOffset) * options.pixelsPerUnit + 0.5
  );
  Point qq(
    (q.x() - options.windowOffset) * options.pixelsPerUnit + 0.5,
    (q.y() - options.windowOffset) * options.pixelsPerUnit + 0.5
  );

  Point vector = qq - pp;
  if (vector.x() == 0 && vector.y() == 0)
  {
    draw((int) pp.y(), (int) pp.x(), pen, colour);
  }
  if (abs(vector.x()) > abs(vector.y()))
  {
    int col = pp.x();
    int dCol = (vector.x() < 0 ? -1 : 1);
    int tCol = qq.x();
    DOUBLE dRow = vector.ySlope() * dCol;
    DOUBLE row = pp.y() - (pp.x() - col) * vector.ySlope();
    for (int j = abs(tCol - col); j >= 0; --j, col += dCol, row += dRow)
      draw((int) row, col, pen, colour);
  }
  else
  {
    int row = pp.y();
    int dRow = (vector.y() < 0 ? -1 : 1);
    int tRow = qq.y();
    DOUBLE dCol = vector.xSlope() * dRow;
    DOUBLE col = pp.x() - (pp.y() - row) * vector.xSlope();
    for (int j = abs(tRow - row); j >= 0; --j, col += dCol, row += dRow)
      draw(row, (int) col, pen, colour);
  }
}


void drawSite(Canvas& c, const Site& s, const int pen, const RGB& colour)
{
  if (s.size == 1) 
    c.draw(s.points[0], s.points[0], pen, colour);
  else for (int i = 1; i < s.size; ++i)
    c.draw(s.points[i-1], s.points[i], pen, colour);
}


/*********************************************************
* Diagram                                                *
*********************************************************/


Element::Element():
  site(-1), shade(1), contour(0)
{};

  
const short Element::softEdge = 1;
const short Element::minorContour = 2;
const short Element::majorContour = 4;
const short Element::bisector = 8;


void putContour(const short level, Element& e1)
{
  e1.contour |= level;
}


void putContour(const short level, Element& e1, Element& e2)
{
  e1.contour |= level;
  e2.contour |= level;
}


/*********************************************************
* Owner                                                  *
*********************************************************/


Owner::Owner():
  primary(-1),
  secondary(-1)
{};


const bool Owner::operator==(const Owner& o) const
{
  return (primary == o.primary && secondary == o.secondary);
}


const bool Owner::operator!=(const Owner& o) const
{
  return !operator==(o);
}


/*********************************************************
* Main                                                   *
*********************************************************/

int main(const int argc, const char* argv[])
{
  if (!std::numeric_limits<DOUBLE>::has_infinity) 
    Complaint(warning) 
      << "Number type does not support infinity. Unexpected results may occur.";

  options.parse(argc, argv);

  if (options.verbose) { cerr << "> Reading sites..."; cerr.flush(); }

  // get sites
  int nrSites; 
  DOUBLE scale;
  if (options.randomSites > 0)
  {
    nrSites = options.randomSites;
    scale = 1.0 / sqrt(nrSites);
  }
  else
  {
    cin >> nrSites;
    if (nrSites < 1) Complaint(userError) << "Too few sites";
    if (nrSites * nrSites > MAXNRSITES) Complaint(userError) << "too many sites";
  }
  if (options.useHighways && nrSites > 1000)
    Complaint(userError) << "too many sites";

  std::vector<Site> sites(nrSites);
  for (int i = 0; i < nrSites; ++i)
  {
    if (options.randomSites > 0)
    {
      if (options.distance->supportedSites() == ROOTEDVECTORTYPE)
        sites[i].generateRandom(1, scale);
      else if (options.distance->supportedSites() == POLYLINETYPE)
        sites[i].generateRandom(2, 2.0 * scale);
      else 
        sites[i].generateRandom(1, 0);
    }
    else
    {
      if (options.distance->supportedSites() == ROOTEDVECTORTYPE) 
        readRootedVector(cin, sites[i]); 
      else 
        cin >> sites[i]; 
    }
    sites[i].setId(i); // <- solution to pass id to highway distance...
    options.distance->verifySite(sites[i]);
  }
  if (options.distance->supports(ROOTEDVECTORTYPE | POLYLINETYPE))
  {
    // check if input consists of segments only; if so, give a warning:
    bool onlySegments = true;
    for (int i = 0; onlySegments && i < nrSites; ++i)
      onlySegments = (sites[i].size == 2);
   
    if (onlySegments) Complaint(warning)
      << "distance function has been implemented for rooted-vector and polyline sites; " 
      << "input will be interpreted as (single-segment) polyline sites";
  }
  if (options.verbose) cerr << "\rNumber of sites: " << nrSites << endl;

  // get highways
  int nrHighways = 0;
  std::vector<Site> highways;
  if (options.useHighways) 
  {
    if (options.verbose) { cerr << "> Reading highways..."; cerr.flush(); }
    if (options.randomSites > 0) 
    {
      nrHighways = min(
        (int) (1.0 + options.randomSites / 12), 
        (int) sqrt(MAXNRSITES / nrSites) * 2);
      scale = 15.0 / sqrt(nrHighways);
    }
    else 
    {
      cin >> nrHighways;
      if (nrHighways < 1) Complaint(userError) << "Missing highways in input file";
      if (nrHighways > 2 * (int) sqrt(MAXNRSITES / nrSites)) 
        Complaint(userError) << "too many highways";
    }
    highways.resize(nrHighways);
    for (int i = 0; i < nrHighways; ++i) 
    {
      if (options.randomSites > 0) 
      {
        highways[i].generateRandomPolyline();
        highways[i].weight = 10;
      }
      else
        cin >> highways[i];
    }
    if (options.verbose) cerr << "\rNumber of highways: " << nrHighways << endl;
    highwayMap = new HighwayMap(sites, highways);
  } 
  
  // set direction for distance comparisons
  int d = (options.order == Options::farthest ? -1 : 1 );

  // scan a grid of squares to precompute viable candidates for closest sites
  if (options.verbose) { cerr << "\r> Preparing pre-scan... "; cerr.flush(); }
  const int gridWidth = (options.width - 1) / cellWidth + 1;
  std::vector<int> prescan;
  int** candidates = newMatrix<int>(gridWidth, gridWidth);
  const int noSite = nrSites;

  DOUBLE lowerBound[nrSites];
  DOUBLE upperBound[nrSites];
  DOUBLE rangeHalfWidth = options.pixelHalfWidth * (cellWidth + 3);

    // compute useful candidate sites for each grid cells
    for (int row = 0; row < gridWidth; ++row)
    {
      if (options.verbose)
        reportProgress(((DOUBLE) row)/gridWidth, "> Pre-scan");
      DOUBLE y = 
        (2.0 * row + 1.0) * cellWidth * options.pixelHalfWidth + options.windowOffset;
      for (int col = 0; col < gridWidth; ++col)
      {
        Point p(
          (2.0 * col + 1.0) * cellWidth * options.pixelHalfWidth + options.windowOffset, 
          y
        );
        candidates[row][col] = prescan.size();
        DOUBLE primUpperBound = INF;
        DOUBLE scndUpperBound = INF;
        DOUBLE maxLowerBound = 0;
        for (int i = 0; i < nrSites; ++i)
        {
          options.distance->getRange(
            p,
            rangeHalfWidth,
            sites[i],
            lowerBound[i],
            upperBound[i]
          );
          keepMax(lowerBound[i], (DOUBLE) 0);
          sites[i].weighDistance(lowerBound[i]);
          sites[i].weighDistance(upperBound[i]);
          keepMin(scndUpperBound, upperBound[i]);
          if (scndUpperBound < primUpperBound) swap(primUpperBound, scndUpperBound);
          keepMax(maxLowerBound, lowerBound[i]);
        }
        if (options.order == Options::first) 
          scndUpperBound = primUpperBound + options.equalityThreshold; 
        for (int i = 0; i < nrSites; ++i)
        {
          if (
            (
              // always calculate distances that are needed to determine 
              // standard contours:
              options.drawStandardContours && 
              lowerBound[i] <= options.contourDistance && 
              options.contourDistance <= upperBound[i]
            ) || (
              options.order != Options::farthest &&
              lowerBound[i] <= scndUpperBound
            ) || (
              options.order == Options::farthest &&
              upperBound[i] >= maxLowerBound
            )
          )
          prescan.push_back(i);
        }
        prescan.push_back(noSite);
      }
    }
    prescan.push_back(noSite); // extra sentinel closing at the end
    if (options.verbose) 
    {
      reportProgress(1, "> Pre-scan");
      cerr << "\rDistance computations eliminated by filter: " 
        << 100.0 * (1.0 - (DOUBLE)
          (cellWidth * cellWidth) * (prescan.size() - 1 - gridWidth * gridWidth) / 
          ((long) options.width * options.width * nrSites)
        ) << '%' << endl;
    }

  // calculate closest sites + distances
  if (options.verbose) { cerr << "\r> Preparing scan...     "; cerr.flush(); }

  // reserve space for two rows with full info
  Sample* samples = new Sample[2 * options.width];
  DOUBLE* distances = new DOUBLE[2 * nrSites * options.width];
  int* layers = new int[2 * nrSites * options.width];
  for (int i = 0; i < 2 * options.width; ++i)
  {
    samples[i].distance = distances; distances += nrSites;
    samples[i].layer    = layers;    layers    += nrSites;
    samples[i].sunlight = 0;
  }
  Sample* scanRow = samples;
  Sample* prevRow = samples + options.width;

  // reserve space for the diagram
  Element** diagram = newMatrix<Element>(options.width+1, options.width+1);

  short boundaryStrength = 0;
  if (options.drawRegionBoundaries) boundaryStrength = Element::bisector;
  else if (options.sunlight) boundaryStrength = Element::softEdge;

  for (int row = 0; row < options.width; ++row)
  {
    if (options.verbose) 
      reportProgress(((DOUBLE) row)/options.width, "> Calculating regions");
    DOUBLE y = (2.0 * row + 1.0) * options.pixelHalfWidth + options.windowOffset;
    int gridRow = row / cellWidth;
    for (int col = 0; col < options.width; ++col)
    {
      Point p((2.0 * col + 1.0) * options.pixelHalfWidth + options.windowOffset, y);
      int gridCol = col / cellWidth;
      Sample& s = scanRow[col];
      Owner& o = s.owner;
      Element& here = diagram[row][col];
      Element& abov = diagram[row+1][col];
      Element& next = diagram[row][col+1];

      if (!options.distanceInfoNeeded && prescan[candidates[gridRow][gridCol]+1] == noSite)
      {
        // exactly one candidate, no distance info needed:
        o.primary   = prescan[candidates[gridRow][gridCol]];
        o.secondary = noSite;
        here.site   = o.primary;
      }
      else
      {
        // calculate distances and keep track of the best two sites:
        o.primary   = noSite; DOUBLE primDistance = INF;
        o.secondary = noSite; DOUBLE scndDistance = INF;
        here.site   = noSite;

        for (
          std::vector<int>::const_iterator sid = prescan.begin() + candidates[gridRow][gridCol]; 
          *sid != noSite; 
          ++sid
        ) 
        {
          Site& site = sites[*sid];
          DOUBLE& distance = s.distance[*sid];
          int& layer = s.layer[*sid];

          distance = (*options.distance)(p, site);
          site.weighDistance(distance);
          layer = (
            options.contourInterval > 0 ?
            (int) floor( (distance-options.contourDistance) / options.contourInterval ) :
            (distance >= options.contourDistance ? 0 : -1)
          );
          // contours are drawn at any 2x2 square (indexed by its upper right corner)
          // that crosses the line; vice versa,
          // if a cell and its left neighbour are on opposite sides, then the squares
          // indexed by that cell and its upper neighbour cross the line;
          // if a cell and its lower neighbour are on opposite sides, then the squares
          // indexed by that cell and its right neighbour cross the line
          // now check these conditions for unit contours:

          // unit circle contours:
          if (options.drawStandardContours)
          { 
            if (col > 0 && (layer < 0) != (scanRow[col-1].layer[*sid] < 0))
              putContour(Element::majorContour, here, abov);
            if (row > 0 && (layer < 0) != (prevRow[col].layer[*sid] < 0))
              putContour(Element::majorContour, here, next);
          }

          // record best distances
          if (distance * d < scndDistance)
          {
            o.secondary = *sid; scndDistance = distance * d;
            if (scndDistance < primDistance)
            {
              swap(o.primary, o.secondary); 
              swap(primDistance, scndDistance);
            }
          }
        }
        // throw away the first-best if not needed:
        if (options.order == Options::nextclosest)
        {
          o.primary = o.secondary; o.secondary = noSite;
        }

        // throw away the second-best if not needed:
        else if (
          options.order != Options::second && 
          scndDistance > primDistance + options.equalityThreshold
        )
          o.secondary = noSite;

        // if there are still two, then lowest-numbered goes first
        if (o.secondary < o.primary) swap(o.primary, o.secondary);

        if (o.primary != noSite && s.distance[o.primary] <= options.cutOffDistance)
        {      
          // choose the site whose colour to use in the diagram:
          here.site = o.primary;
          if (o.secondary != noSite &&
#ifdef CHECKERBOARD
            (row / strokeWidth[hatchPen] + col / strokeWidth[hatchPen]) % 2 == 0
#else
            ((row + col) / strokeWidth[hatchPen]) % 2 == 0
#endif
          )
            here.site = o.secondary;
          if (options.shadeInnerAreas && s.distance[here.site] <= options.contourDistance)
            here.shade = 0.7529;
          if (options.halfBrightDistance > 0)
            here.shade *= pow(0.5, s.distance[here.site] / options.halfBrightDistance) * 1.14;
        }

        // let N, NE, E, etc. be the North, East, North East neighbours;
        // let dN, dNE, dE, etc. be distance of current location - their distance;
        if (options.sunlight && row > 0 && col > 0 && col < options.width-1 && o.primary != noSite)
        {
          DOUBLE dSE = (s.distance[o.primary] - prevRow[col+1].distance[o.primary]) * options.pixelsPerUnit; 
          DOUBLE dS  = (s.distance[o.primary] - prevRow[col]  .distance[o.primary]) * options.pixelsPerUnit;
          DOUBLE dSW = (s.distance[o.primary] - prevRow[col-1].distance[o.primary]) * options.pixelsPerUnit;
          DOUBLE dW  = (s.distance[o.primary] - scanRow[col-1].distance[o.primary]) * options.pixelsPerUnit;

          // for each point, determine brightness by taking average over all 12 possible incident
          // triangles in a triangulation of the grid
          // brightness ~ inner product of triangle's normal and the sun. 
          DOUBLE sunlight;
          DOUBLE dotProduct;
          Sample& SE = prevRow[col+1];
          Sample& S  = prevRow[col];
          Sample& SW = prevRow[col-1];
          Sample& W  = scanRow[col-1];

          // triangle with S, SE has normal (dS-dSE, dS, 1)
          sunlight = safeCosXY1WithSun(dS-dSE, dS);
          s.sunlight += sunlight; S.sunlight += sunlight; SE.sunlight += sunlight;
          
          // triangle with S, SW has normal (dSW-dS, dS, 1)
          sunlight = safeCosXY1WithSun(dSW-dS, dS);
          s.sunlight += sunlight; S.sunlight += sunlight; SW.sunlight += sunlight;

          // triangle with S, W has normal (dW, dS, 1)
          sunlight = safeCosXY1WithSun(dW, dS);
          s.sunlight += sunlight; S.sunlight += sunlight; W.sunlight += sunlight;

          // triangle with SW, W has normal (dW, dSW-dW, 1)
          sunlight = safeCosXY1WithSun(dW, dSW-dW);
          s.sunlight += sunlight; SW.sunlight += sunlight; W.sunlight += sunlight;
        }
      }

      // check the conditions for region contours (bisectors)
      if (col > 0 && o != scanRow[col-1].owner) putContour(boundaryStrength, here, abov);
      if (row > 0 && o != prevRow[col].owner) putContour(boundaryStrength, here, next);

      // check the conditions for regular distance contours
      if (options.contourInterval > 0)
      {
        if (col > 0 && 
          here.site == diagram[row][col-1].site && 
          s.layer[here.site] != scanRow[col-1].layer[here.site]
        )
        {
          short level = 
            xyInDifferentBracketsOfz(s.layer[here.site], scanRow[col-1].layer[here.site], 5) ?
            Element::majorContour :
            Element::minorContour;
          putContour(level, here);
          if (abov.site == here.site) putContour(level, abov);
        }
        if (row > 0 && 
          here.site == diagram[row-1][col].site && 
          s.layer[here.site] != prevRow[col].layer[here.site]
        )
        {
          short level = 
            xyInDifferentBracketsOfz(s.layer[here.site], prevRow[col].layer[here.site], 5) ?
            Element::majorContour :
            Element::minorContour;
          putContour(level, here);
          if (next.site == here.site) putContour(level, next);
        }
      }
    }
    if (options.sunlight && row > 0)
      for (int col = 0; col < options.width; ++col)
      {
        DOUBLE& sunlight = prevRow[col].sunlight;
        diagram[row][col].shade *= 0.8 + 0.02 * sunlight;
        // prepare for reuse as scan row in next iteration:
        sunlight = 0;
      }
    swap(scanRow, prevRow);
  }
  if (options.verbose) reportProgress(1, "> Calculating regions");

  if (options.verbose) { cerr << "\r> Choosing colours...       "; cerr.flush(); }

  bool** adjacent = newMatrix<bool>(nrSites+1, nrSites+1);
  for (int i = 0; i < nrSites; ++i)
    for (int j = 0; j < nrSites; ++j)
      adjacent[i][j] = false;

  if (options.colourAssignment != Options::inputOrder)
  {
    for (int row = 1; row < options.width; ++row)
      for (int col = 1; col < options.width; ++col)
      {
        adjacent[diagram[row][col].site][diagram[row-1][col].site] = true;
        adjacent[diagram[row][col].site][diagram[row][col-1].site] = true;
        adjacent[diagram[row][col].site][diagram[row-1][col-1].site] = true;
      }
    for (int i = 0; i < nrSites; ++i)
    {
      for (int j = 0; j < nrSites; ++j)
        adjacent[i][j] = adjacent[i][j] || adjacent[j][i];
      adjacent[i][i] = false;
    }
  }
  RGB* colour = chooseColours(options, nrSites, adjacent);

  // prepare canvas
  if (options.verbose) { cerr << "\r> Preparing canvas...   "; cerr.flush(); }
  Canvas canvas(options.width, options.width);

  // draw regions
  for (int row = 0; row < options.width; ++row)
  {
    if (options.verbose) 
      reportProgress(((DOUBLE) row)/options.width, "> Drawing regions");
    for (int col = 0; col < options.width; ++col)
    {
      Element& here = diagram[row][col];
      canvas.shade(row, col, colour[here.site]);
      canvas.shade(row, col, here.shade);
    }
  }
  if (options.verbose) reportProgress(1, "> Drawing regions");

  // draw contours and bisectors
  if (options.lineWork() || options.sunlight)
  {
    for (int row = 0; row < options.width; ++row)
    {
      if (options.verbose) 
        reportProgress(((DOUBLE) row)/options.width, "> Drawing contours");
      for (int col = 0; col < options.width; ++col)
      {        
        short contour = diagram[row][col].contour;
        if (contour == 0) continue;
        if (contour & Element::bisector)
          canvas.draw(row, col, bisectorPen, RGB::black);
        if (contour & Element::majorContour)
          canvas.draw(row, col, contourPen, RGB::black);
        if (contour & Element::minorContour)
          canvas.shade(row, col, contourPen, RGB::lightGrey);
        if (contour & Element::softEdge)
          canvas.shade(row, col, softBisectorPen, RGB::faintGrey);
      }
    }
    if (options.verbose) reportProgress(1, "> Drawing contours");
  }

  // draw highways
  if (nrHighways > 0)
  {
    if (options.verbose) { cerr << "\r> Drawing highways...        "; cerr.flush(); }

    // draw the highways
    for (int i = 0; i < nrHighways; ++i)
      drawSite(canvas, highways[i], sitePen, RGB::black);
    for (int i = 0; i < nrHighways; ++i)
      drawSite(canvas, highways[i], bisectorPen, RGB::grey);
    // highwayMap->showNodes(canvas); // for debugging only
  }

  // draw sites
  if (options.drawSites)
  {
    if (options.verbose) { cerr << "\r> Drawing sites...           "; cerr.flush(); }

    // draw the sites
    for (int i = 0; i < nrSites; ++i)
    {
      // arrow:
      Point arrowHead(sites[i].points[0] + sites[i].vectors[0]);
      canvas.draw(sites[i].points[0], arrowHead, arrowPen, RGB::black);     
      canvas.draw(arrowHead, arrowHead, arrowHeadPen, RGB::black);
 
      // points/polylines:
      if (options.nrColours > 1)
      {
        drawSite(canvas, sites[i], siteOutlinePen, RGB::black);
        if (options.unicolourSites)
          drawSite(canvas, sites[i], sitePen, RGB::white);
        else
          drawSite(canvas, sites[i], sitePen, colour[i]);
      }
      else
        drawSite(canvas, sites[i], sitePen, RGB::black);
    }
  }

  // mark the origin
  if (options.markOrigin)
  {
    for (
      int row = options.width/2 - 5 * strokeWidth[contourPen];
      row < options.width/2 + 5 * strokeWidth[contourPen];
      ++row)
    {
      canvas.draw(row, options.width/2, markOutlinePen, RGB::white);     
      canvas.draw(options.width/2, row, markOutlinePen, RGB::white);     
    }
    for (
      int row = options.width/2 - 5 * strokeWidth[contourPen];
      row < options.width/2 + 5 * strokeWidth[contourPen];
      ++row)
    {
      canvas.draw(row, options.width/2, contourPen, RGB::black);     
      canvas.draw(options.width/2, row, contourPen, RGB::black);     
    }
  }

  if (options.verbose) { cerr << "\r> Writing bitmap file...     "; cerr.flush(); }
  // output bitmap
  BitmapFile bmp(cout, options.width, options.width);
  for (int row = 0; row < options.width; ++row)
    for (int col = 0; col < options.width; ++col)
      bmp.writePixel(canvas(row,col));
  if (options.verbose) cerr << "\r> All done!              " << endl;
}
